# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Provide support for legacy UserInfo object."""

from flask import current_app, has_request_context, request, session
from flask_login import UserMixin
from werkzeug.datastructures import CallbackDict, CombinedMultiDict

from invenio_ext.cache import cache

__all__ = ('UserInfo', )

CFG_USER_DEFAULT_INFO = {
    'remote_ip': '',
    'remote_host': '',
    'referer': '',
    'uri': '',
    'agent': '',
    'uid': -1,
    'nickname': '',
    'given_names': '',
    'family_name': '',
    'email': '',
    'group': [],
    'guest': '1',
    'session': None,
    'precached_permitted_restricted_collections': [],
    'precached_allowed_deposition_types': [],
    'precached_useloans': False,
    'precached_usegroups': False,
    'precached_usemessages': False,
    'precached_viewsubmissions': False,
    'precached_useapprove': False,
    'precached_useadmin': False,
    'precached_viewclaimlink': False,
    'precached_usepaperclaim': False,
    'precached_usepaperattribution': False,
    'precached_canseehiddenmarctags': False,
    'precached_sendcomments': False,
}


class UserInfo(CombinedMultiDict, UserMixin):

    """Provide legacy implementation.

    Methods that Flask-Login and Invenio 1.x expect user objects to have.
    """

    def __init__(self, uid=None, force=False):
        """Retrieve information about user."""
        def on_update(self):
            """Change own status when the user info is modified."""
            self.modified = True

        self.modified = False
        self.uid = uid
        self.req = self._get_request_info()
        acc = {}

        if uid is not None and uid > 0:
            data = self._login(uid, force)
            acc = self._precache(data, force)
        else:
            data = self._create_guest()

        self.info = CallbackDict(data, on_update)
        # FIXME remove req after everybody start using flask request.
        CombinedMultiDict.__init__(self, [self.req, self.info, acc,
                                          dict(CFG_USER_DEFAULT_INFO)])
        self.save()

    def get_key(self):
        """Generate key for caching user information."""
        key = 'current_user::' + str(self.uid)
        return key

    def get_acc_key(self):
        """Generate key for caching authorizations."""
        remote_ip = str(request.remote_addr) if has_request_context() else '0'
        return 'current_user::' + str(self.uid) + '::' + remote_ip

    def save(self):
        """Save modified data permanently for logged users."""
        if not self.is_guest and self.modified:
            timeout = current_app.config.get(
                'CFG_WEBSESSION_EXPIRY_LIMIT_DEFAULT', 0) * 3600
            cache.set(self.get_key(), dict(self.info),
                      timeout=timeout)

    def reload(self):
        """Reload user login information and saves them."""
        data = self._login(self.uid, force=True)
        acc = self._precache(data, force=True)
        self.info.update(data)
        CombinedMultiDict.__init__(self, [self.req, self.info, acc,
                                          dict(CFG_USER_DEFAULT_INFO)])
        self.save()

    def update_request_info(self):
        """Update request information."""
        self.req = self._get_request_info()

    def _get_request_info(self):
        """Get request information."""
        # FIXME: we should support IPV6 too. (hint for FireRole)
        data = {}
        if has_request_context():
            data['remote_ip'] = request.remote_addr or ''
            data['remote_host'] = request.environ.get('REMOTE_HOST', '')
            data['referer'] = request.referrer
            data['uri'] = request.environ['PATH_INFO'] or ''
            data['agent'] = request.headers.get('User-Agent', 'N/A')
        return data

    def _create_guest(self):
        # Minimal information about user.
        return {'settings': {}, 'id': 0, 'uid': 0}

    def _login(self, uid, force=False):
        """Get account information about currently logged user from database.

        Should raise an exception when session.uid is not valid User.id.
        """
        data = cache.get(self.get_key())
        if not force and data is not None:
            return data

        from invenio_accounts.models import User
        data = {}

        try:
            user = User.query.get(uid)
            data['id'] = data['uid'] = user.id or -1
            data['nickname'] = user.nickname or ''
            data['given_names'] = user.given_names or ''
            data['family_name'] = user.family_name or ''
            data['email'] = user.email or ''
            data['note'] = user.note or ''
            data['group'] = map(
                lambda x: x.group.name, getattr(user, 'groups', [])
            )
            data.update(user.settings or {})
            data['settings'] = user.settings or {}
            data['guest'] = str(int(user.guest))  # '1' or '0'
            self.modified = True
        except Exception:
            data = self._create_guest()

        return data

    def _precache(self, info, force=False):
        """Calculate permissions for user actions.

        FIXME: compatibility layer only !!!
        """
        try:
            from invenio_accounts.models import User
        except ImportError:
            return {}

        CFG_BIBAUTHORID_ENABLED = current_app.config.get(
            'CFG_BIBAUTHORID_ENABLED', False)
        # get authorization key
        acc_key = self.get_acc_key()
        acc = cache.get(acc_key)
        if not force and acc_key is not None and acc is not None:
            return acc

        # FIXME: acc_authorize_action should use flask request directly
        user_info = info
        user_info.update(self.req)
        user = User.query.get(user_info['uid'])


        data = {}
        data['precached_useadmin'] = user.has_admin_role
        data['precached_usesuperadmin'] = user.has_super_admin_role

        try:
            from invenio_search.utils import \
                get_permitted_restricted_collections
            data['precached_permitted_restricted_collections'] = \
                get_permitted_restricted_collections(user_info)
        except Exception:
            current_app.logger.exception(
                'Permitted restricted collections were not loaded.'
            )

        try:
            from invenio_deposit.cache import \
                get_authorized_deposition_types
            data['precached_allowed_deposition_types'] = \
                get_authorized_deposition_types(user_info)
        except Exception:
            current_app.logger.exception(
                'Allowed deposition types were not loaded.'
            )

        usepaperclaim = False
        usepaperattribution = False
        viewclaimlink = False
        try:
            from invenio_access.engine import acc_authorize_action
            from invenio_access.control import acc_get_role_id, \
                acc_is_user_in_role
            data['precached_useloans'] = acc_authorize_action(
                user_info, 'useloans')[0] == 0
            data['precached_usegroups'] = acc_authorize_action(
                user_info, 'usegroups')[0] == 0
            data['precached_usemessages'] = acc_authorize_action(
                user_info, 'usemessages')[0] == 0
            data['precached_canseehiddenmarctags'] = acc_authorize_action(
                user_info, 'runbibedit')[0] == 0

            if (CFG_BIBAUTHORID_ENABLED and acc_is_user_in_role(
                    user_info, acc_get_role_id("paperclaimviewers"))):
                usepaperclaim = True

            if (CFG_BIBAUTHORID_ENABLED and acc_is_user_in_role(
                    user_info, acc_get_role_id("paperattributionviewers"))):
                usepaperattribution = True
        except Exception:
            current_app.logger.exception(
                "Access control module is broken."
            )

        viewlink = False
        try:
            viewlink = session['personinfo']['claim_in_process']
        except (KeyError, TypeError):
            pass

        if (current_app.config.get('CFG_BIBAUTHORID_ENABLED') and
                usepaperattribution and viewlink):
            viewclaimlink = True

#       if (CFG_BIBAUTHORID_ENABLED
#               and ((usepaperclaim or usepaperattribution)
#               and acc_is_user_in_role(
#                   data, acc_get_role_id("paperattributionlinkviewers")))):
#           viewclaimlink = True

        data['precached_viewclaimlink'] = viewclaimlink
        data['precached_usepaperclaim'] = usepaperclaim
        data['precached_usepaperattribution'] = usepaperattribution

        timeout = current_app.config.get(
            'CFG_WEBSESSION_EXPIRY_LIMIT_DEFAULT', 0) * 3600
        cache.set(acc_key, data,
                  timeout=timeout)
        return data

    def is_authenticated(self):
        """Check if user is authenticated."""
        return not self.is_guest

    def is_authorized(self, name, **kwargs):
        """Check if user is authorized."""
        from invenio_access.engine import acc_authorize_action
        return acc_authorize_action(self, name)[0] == 0

    def is_active(self):
        """Check if user is active."""
        return not self.is_guest

    def is_confirmed(self):
        """Return true if accounts has been confirmed."""
        return self['note'] == "1"

    @property
    def is_guest(self):
        """Check if user is guest."""
        return True if self['email'] == '' else False

    @property
    def is_admin(self):
        """Check if user is admin."""
        return self.get('precached_useadmin', False)

    @property
    def is_super_admin(self):
        """Check if user is super admin."""
        return self.get('precached_usesuperadmin', False)

    def get_id(self):
        """Get user id."""
        return self.get('id', -1)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Invenio module that provides integration with Flask extensions."""

import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

requirements = [
    # Cerberus>=0.7.1 api changes and is not yet supported
    'Cerberus>=0.7,<0.7.1',
    'Flask-Admin>=1.1.0',
    'Flask-Assets>=0.10',
    'Flask-Babel>=0.9',
    'Flask-Breadcrumbs>=0.2',
    'Flask-Cache>=0.12',
    'Flask-Collect>=1.1.1',
    'Flask-Email>=1.4.4',
    'Flask-Gravatar>=0.4.2',
    'Flask-IIIF>=0.2.0',
    'Flask-Login>=0.2.7',
    'Flask-Menu>=0.2',
    'Flask-OAuthlib>=0.6.0,<0.7',  # quick fix for issue #2158
    'Flask-Principal>=0.4',
    'Flask-RESTful>=0.2.12',
    'Flask-Registry>=0.2',
    'Flask-SQLAlchemy>=2.0',
    'Flask-Script>=2.0.5',
    'Flask-WTF>=0.10.2',
    'Flask>=0.10.1',
    'MySQL-python>=1.2.5',
    'backports.lzma>=0.0.3',
    'celery>=3.1.8',
    'elasticsearch>=1.3.0',
    'fixture>=1.5',
    'intbitset>=2.0.0',
    'invenio-base>=0.3.0',
    'invenio-celery>=0.1.0',
    'invenio-collections>=0.1.2',
    'invenio-utils>=0.1.1',
    'lxml>=3.3',
    # FIXME new oauthlib release after 0.7.2 has some compatible problems with
    # the used Flask-Oauthlib version.
    'oauthlib==0.7.2',
    'passlib>=1.6.2',
    'python-dateutil>=1.5',
    'raven>=5.0.0',
    'redis>=2.8.0',
    'requests>=2.4.0',
    'six>=1.7.2',
    'sqlalchemy-utils[encrypted]>=0.31.0',
]

test_requirements = [
    'coverage>=4.0.0',
    'Flask-SSO>=0.2',
    'httpretty>=0.8.10',
    'invenio-accounts>=0.1.2',
    'invenio-oauth2server>=0.1.1',
    'invenio-testing>=0.1.1',
    'mock>=1.0.0',
    'pytest>=2.8.0',
    'pytest-cov>=2.1.0',
    'pytest-isort>=0.1.0',
    'pytest-pep8>=1.0.6',
    'setuptools>=17.1',
]


class PyTest(TestCommand):

    """PyTest Test."""

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        """Init pytest."""
        TestCommand.initialize_options(self)
        self.pytest_args = []
        try:
            from ConfigParser import ConfigParser
        except ImportError:
            from configparser import ConfigParser
        config = ConfigParser()
        config.read('pytest.ini')
        self.pytest_args = config.get('pytest', 'addopts').split(' ')

    def finalize_options(self):
        """Finalize pytest."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Run tests."""
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('invenio_ext', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='invenio-ext',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='invenio TODO',
    license='GPLv2',
    author='CERN',
    author_email='info@invenio-software.org',
    url='https://github.com/inveniosoftware/invenio-ext',
    packages=[
        'invenio_ext',
    ],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=requirements,
    extras_require={
        'docs': [
            'Sphinx>=1.3',
            'sphinx_rtd_theme>=0.1.7'
        ],
        'tests': test_requirements
    },
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
        'Development Status :: 1 - Planning',
    ],
    tests_require=test_requirements,
    cmdclass={'test': PyTest},
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
"""
Configuration class for `jinja2.bccache`.

.. code-block:: python

    JINJA2_BCCACHE = False
    JINJA2_BCCACHE_SIZE = -1
    JINJA2_BCCACHE_AUTO_RELOAD = False
    JINJA2_BCCACHE_CLIENT = 'invenio_ext.cache:cache'
    JINJA2_BCCACHE_PREFIX = 'jinja2::bccache::'
    JINJA2_BCCACHE_TIMEOUT = None
    JINJA2_BCCACHE_IGNORE_CACHE_ERRORS = True

"""

from jinja2.bccache import MemcachedBytecodeCache
from werkzeug.utils import import_string


class BytecodeCacheWithConfig(MemcachedBytecodeCache):

    """A bytecode cache that uses application config for initialization."""

    def __init__(self, app):
        """Initialize `BytecodeCache` from application config."""
        client_string = app.config.get('JINJA2_BCCACHE_CLIENT',
                                       'invenio_ext.cache:cache')
        client = import_string(client_string)
        prefix = app.config.get('JINJA2_BCCACHE_PREFIX', 'jinja2::bccache::')
        timeout = app.config.get('JINJA2_BCCACHE_TIMEOUT', None)
        ignore_memcache_errors = app.config.get(
            'JINJA2_BCCACHE_IGNORE_CACHE_ERRORS', True)
        super(self.__class__, self).__init__(
            client, prefix=prefix, timeout=timeout,
            ignore_memcache_errors=ignore_memcache_errors)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Test *Flask-SSO* integration."""

from __future__ import absolute_import

import sys
import unittest
from contextlib import contextmanager

from flask import request, request_started
from flask_login import current_user

from invenio_testing import InvenioTestCase

try:
    from invenio_ext.sso import setup_app
    has_sso = True
except ImportError:
    has_sso = False


class TestSSO(InvenioTestCase):

    """Test extension itegration."""

    @unittest.skipUnless(has_sso, 'Flask-SSO is not installed')
    def test_login_handler(self):
        """Test login handler."""
        self.app = setup_app(self.app)

        @contextmanager
        def request_environ_set(app, data):

            def handler(sender, **kwargs):
                for (k, v) in data.items():
                    request.environ[k] = v

            with request_started.connected_to(handler, app):
                yield

        def run(data, expected_data):
            with request_environ_set(self.app, data):
                with self.app.test_client() as c:
                    c.get(self.app.config['SSO_LOGIN_URL'])
                    current_user['email'] == expected_data['email']
                    current_user['group'] == expected_data['group']

        data = {
            'ADFS_GROUP': ('CERN Registered;'
                           'project-invenio-devel;'
                           'cern-personnel'),
            'ADFS_LOGIN': 'admin',
            'ADFS_EMAIL': self.app.config['CFG_SITE_ADMIN_EMAIL'],
        }
        expected_data = {
            'email': self.app.config['CFG_SITE_ADMIN_EMAIL'],
            'group': [
                'project-invenio-devel (Group)', 'CERN Registered (Group)',
                'cern-personnel (Group)'
            ],
        }

        # FIXME mock user table
        from invenio_ext.sqlalchemy import db
        from invenio_accounts.models import User
        admin = User.query.get(1)
        old_settings = admin.settings
        admin.settings = {'login_method': 'SSO'}
        db.session.merge(admin)
        db.session.commit()

        run(data, expected_data)

        admin.settings = old_settings
        db.session.merge(admin)
        db.session.commit()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Flask-Cache backend for session.

Configuration variables for Flask-Cache backend.

=============================== ===============================================
`SESSION_BACKEND_CACHE`         Configured *Flask-Cache* object.
                                **Default:** ``invenio_ext.cache:cache``
`SESSION_BACKEND_CACHE_PREFIX`  Prefix for keys stored in cache.
                                **Default:** ``session::``
`SESSION_BACKEND_CACHE_TIMEOUT` Default cache timeout. **Default:** ``3600``
=============================== ===============================================
"""

from datetime import timedelta

from flask import current_app
from flask.helpers import locked_cached_property
from werkzeug.utils import import_string

from ..storage import SessionStorage


class Storage(SessionStorage):

    """Implement session cache storage."""

    @locked_cached_property
    def cache(self):
        """Return cache storage."""
        return import_string(current_app.config.get(
            'SESSION_BACKEND_CACHE', 'invenio_ext.cache:cache'))

    @locked_cached_property
    def key(self):
        """Return cache key prefix."""
        return current_app.config.get('SESSION_BACKEND_CACHE_PREFIX',
                                      'session::')

    @locked_cached_property
    def timeout(self):
        """Return cache engine timeout."""
        return current_app.config.get('SESSION_BACKEND_CACHE_TIMEOUT', 3600)

    def set(self, name, value, timeout=None):
        """Store value in cache."""
        timeout = timeout if timeout is not None else self.timeout
        # Convert datetime.timedeltas to seconds.
        if isinstance(timeout, timedelta):
            timeout = timeout.seconds + timeout.days * 24 * 3600
        self.cache.set(self.key + name, value, timeout=timeout)

    def get(self, name):
        """Return value from cache."""
        return self.cache.get(self.key + name)

    def delete(self, name):
        """Delete key from cache."""
        self.cache.delete(self.key + name)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
errorlib database models.
"""

# General imports.
from datetime import datetime

from invenio_base.globals import cfg
from invenio_ext.sqlalchemy import db


def _is_pow_of_2(n):
    """
    Return True if n is a power of 2
    """
    while n > 1:
        if n % 2:
            return False
        n = n / 2
    return True


class HstEXCEPTION(db.Model):
    """Represents a HstEXCEPTION record."""
    __tablename__ = 'hstEXCEPTION'
    id = db.Column(db.Integer(15, unsigned=True), nullable=False,
                   primary_key=True, autoincrement=True)
    name = db.Column(db.String(50), nullable=False)
    filename = db.Column(db.String(255), nullable=True)
    line = db.Column(db.Integer(9), nullable=True)
    last_seen = db.Column(db.DateTime, nullable=False,
                          server_default='1900-01-01 00:00:00', index=True)
    last_notified = db.Column(db.DateTime, nullable=False,
                              server_default='1900-01-01 00:00:00', index=True)
    counter = db.Column(db.Integer(15), nullable=False,
                        server_default='0')
    total = db.Column(db.Integer(15), nullable=False,
                      server_default='0', index=True)

    __table_args__ = (db.Index('name', name, filename, line, unique=True),
                      db.Model.__table_args__)

    @classmethod
    def get_or_create(cls, name, filename, line):
        """Finds or create exception log."""
        try:
            log = cls.query.filter_by(name=name, filename=filename,
                                      line=line).one()
            delta = datetime.datetime.now() - log.last_notified
            reset_counter = (delta.seconds + delta.days * 86400) >= \
                cfg['CFG_ERRORLIB_RESET_EXCEPTION_NOTIFICATION_COUNTER_AFTER']
            counter = 1 if reset_counter else log.counter + 1
            log.update({'last_notified': db.func.now(),
                        'counter': counter,
                        'total': log.total + 1}, synchronize_settion=False)
            db.session.add(log)
        except:
            log = HstEXCEPTION(name=name,
                               filename=filename,
                               line=line,
                               last_seen=datetime.now(),
                               last_notified=datetime.now(),
                               counter=1,
                               total=1)
            db.session.add(log)
        try:
            db.session.commit()
        except:
            db.session.rollback()
        return log

    @property
    def exception_should_be_notified(self):
        return _is_pow_of_2(self.counter)

    @property
    def pretty_notification_info(self):
        return ("This exception has already been seen %s times\n    "
                "last time it was seen: %s\n    "
                "last time it was notified: %s\n" % (
                    self.total,
                    self.last_seen.strftime("%Y-%m-%d %H:%M:%S"),
                    self.last_notified.strftime("%Y-%m-%d %H:%M:%S")))

    @classmethod
    def get_pretty_notification_info(cls, name, filename, line):
        """
        Return a sentence describing when this exception was already seen.
        """
        try:
            return cls.query.filter_by(
                name=name, filename=filename, line=line
            ).one().pretty_notification_info
        except:
            return "It is the first time this exception has been seen.\n"


__all__ = ['HstEXCEPTION']



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Initialize and configure *Flask-SSO* extension."""

import re

from flask import flash, redirect
from flask_sso import SSO

#: Tunable list of settings to be hidden.
#: e.g.: CFG_EXTERNAL_AUTH_HIDDEN_SETTINGS = ('auth', 'respccid', 'personid')
CFG_EXTERNAL_AUTH_HIDDEN_SETTINGS = ()

#: Tunable list of groups to be hidden.
CFG_EXTERNAL_AUTH_HIDDEN_GROUPS = (
    'All Exchange People',
    'CERN Users',
    'cern-computing-postmasters',
    'cern-nice2000-postmasters',
    'CMF FrontEnd Users',
    'CMF_NSC_259_NSU',
    'Domain Users',
    'GP Apply Favorites Redirection',
    'GP Apply NoAdmin',
    'info-terminalservices',
    'info-terminalservices-members',
    'IT Web IT',
    'NICE Deny Enforce Password-protected Screensaver',
    'NICE Enforce Password-protected Screensaver',
    'NICE LightWeight Authentication WS Users',
    'NICE MyDocuments Redirection (New)',
    'NICE Profile Redirection',
    'NICE Terminal Services Users',
    'NICE Users',
    'NICE VPN Users',
)

#: Tunable list of regexps of groups to be hidden.
CFG_EXTERNAL_AUTH_HIDDEN_GROUPS_RE = (
    re.compile(r'Users by Letter [A-Z]'),
    re.compile(r'building-[\d]+'),
    re.compile(r'Users by Home CERNHOME[A-Z]'),
)

#: Default attribute map
SSO_ATTRIBUTE_MAP = {
    'ADFS_AUTHLEVEL': (False, 'authlevel'),
    'ADFS_GROUP': (True, 'group'),
    'ADFS_LOGIN': (True, 'nickname'),
    'ADFS_ROLE': (False, 'role'),
    'ADFS_EMAIL': (True, 'email'),
    'ADFS_IDENTITYCLASS': (False, 'external'),
    'HTTP_SHIB_AUTHENTICATION_METHOD': (False, 'authmethod'),
}

sso = SSO()


def setup_app(app):
    """Setup SSO extension."""
    app.config['CFG_EXTERNAL_AUTH_USING_SSO'] = True
    app.config.setdefault('SSO_ATTRIBUTE_MAP', SSO_ATTRIBUTE_MAP)
    sso.init_app(app)

    def fetch_groups(groups):
        groups = groups.split(
            app.config.get('CFG_EXTERNAL_AUTH_SSO_GROUPS_SEPARATOR', ';'))
        # Filtering out uncomfortable groups
        groups = [group for group in groups if group not in app.config.get(
            'CFG_EXTERNAL_AUTH_HIDDEN_GROUPS',
            CFG_EXTERNAL_AUTH_HIDDEN_GROUPS)]
        for regexp in app.config.get('CFG_EXTERNAL_AUTH_HIDDEN_GROUPS_RE',
                                     CFG_EXTERNAL_AUTH_HIDDEN_GROUPS_RE):
            for group in groups:
                if regexp.match(group):
                    groups.remove(group)
        return dict(map(lambda x: (x, '@' in x and x + ' (Mailing list)'
                                   or x + ' (Group)'), groups))

    def fetch_external(external):
        return '0' if external in ('CERN Registered', 'CERN Shared') else '1'

    @sso.login_handler
    def login_callback(user_info):
        """Login user base on SSO context (create one if necessary).

        Function should not raise an exception if `user_info` is not valid
        or `User` was not found in database.
        """
        from invenio_accounts.models import User
        from invenio_ext.login import (authenticate, login_redirect,
                                       current_user)
        from invenio_ext.sqlalchemy import db

        user_info['group'] = fetch_groups(user_info['group']).values()
        user_info['external'] = fetch_external(user_info.get('external'))
        try:
            auth = authenticate(user_info['email'], login_method='SSO')
            if auth is None:
                user = User()
                user.nickname = user_info['nickname']
                user.email = user_info['email']
                user.password = ''
                user.settings = {'login_method': 'SSO'}
                db.session.add(user)
                db.session.commit()
                auth = authenticate(user_info['email'], login_method='SSO')
                if auth is None:
                    return redirect('/')

            current_user.info['group'] = current_user.get('group', []) + \
                user_info['group']
            current_user.save()
        except:
            flash('Problem with login (%s)' % (str(user_info)), 'error')
            return redirect('/')

        return login_redirect()

    return app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement ZlibPickable column type."""

from sqlalchemy.types import LargeBinary, TypeDecorator

from invenio_utils.serializers import ZlibPickle


class PickleBinary(TypeDecorator):

    """Implement ZlibPickable column type."""

    impl = LargeBinary

    def process_bind_param(self, value, dialect):
        """Dump data to column using ZlibPickle."""
        if value is not None:
            value = ZlibPickle.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        """Load ZlibPickled data from column."""
        if value is not None:
            value = ZlibPickle.loads(value)
        return value



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Initialization and configuration for *Flask-Restful*."""

import re
import six
import warnings

from datetime import date
from dateutil import parser
from dateutil.tz import tzlocal, tzutc
from flask import request, session
from flask.ext import restful
from flask_restful import fields
from flask_registry import ModuleAutoDiscoveryRegistry
from functools import wraps
from cerberus import Validator

error_codes = dict(
    validation_error=10,
)
"""
Available error codes for REST API.
"""

# errors and codes when validating JSON data concerning restful APIs
validation_errors = dict(
    INCORRECT_TYPE=dict(
        error_code=1,
        error_mesg="An Attribute has incorrect type according to schema"
    ),
    MISSING_FROM_USER_INPUT=dict(
        error_code=2,
        error_mesg="Input is missing a required field"
    ),
    NON_EXISTING_TO_SCHEMA=dict(
        error_code=3,
        error_mesg="Input contains a field that does not exist in schema"
    ),
    NO_UTC_ISO_FORMAT=dict(
        error_code=4,
        error_mesg=("Input contains datetime attribute "
                    "that is not in utc iso format")
    ),
    DATETIME_PARSE_ERROR=dict(
        error_code=5,
        error_mesg="Input contains datetime attribute that cannot be parsed"
    ),
    VALUE_OUT_OF_BOUNDS=dict(
        error_code=6,
        error_mesg="Input contains an attribute with an out of bounds value"
    ),
    INCORRECT_ELEMENT_TYPE_IN_DATASTRUCTURE=dict(
        error_code=7,
        error_mesg="Elements in data structure have incorrect type"),
)


class RESTValidator(Validator):

    """Validator for restful Api."""

    def _validate_utciso(self, utciso, field, value):
        """Validate UTC ISO format."""
        try:
            dt = parser.parse(value)
            if dt.tzinfo != tzutc():
                self._error(field, "not in utc iso format")
        except Exception:
            self._error(field, "cannot parse date-time")

    def get_errors(self):
        """Transform cerberus validator errors to a list of dictionaries.

        Example::

            {
                "code": c,
                "message": "a message",
                "field": f
            }
        """
        found_errors = []
        all_errors = self.errors

        for key in all_errors:
            if isinstance(all_errors[key], str):
                msg_error = all_errors[key]

                if re.match(
                    "must be of (string|integer|float|boolean|list) type",
                    msg_error
                ):
                    error_to_append = dict(
                        code=validation_errors['INCORRECT_TYPE']['error_code'],
                        message=(
                            validation_errors['INCORRECT_TYPE']['error_mesg']
                            + ": " + "'" + key + "' " + msg_error
                        ),
                        field=key
                    )
                    found_errors.append(error_to_append)

                elif msg_error == "unknown field":
                    error_to_append = dict(
                        code=(validation_errors['NON_EXISTING_TO_SCHEMA']
                              ['error_code']),
                        message=(validation_errors['NON_EXISTING_TO_SCHEMA']
                                 ['error_mesg']),
                        field=key
                    )
                    found_errors.append(error_to_append)

                elif msg_error == "required field":
                    error_to_append = dict(
                        code=(validation_errors['MISSING_FROM_USER_INPUT']
                              ['error_code']),
                        message=(validation_errors['MISSING_FROM_USER_INPUT']
                                 ['error_mesg']),
                        field=key
                    )
                    found_errors.append(error_to_append)

                elif msg_error == "not in utc iso format":
                    error_to_append = dict(
                        code=(validation_errors['NO_UTC_ISO_FORMAT']
                              ['error_code']),
                        message=(validation_errors['NO_UTC_ISO_FORMAT']
                                 ['error_mesg']),
                        field=key
                    )
                    found_errors.append(error_to_append)
                elif msg_error == "cannot parse date-time":
                    error_to_append = dict(
                        code=(validation_errors['DATETIME_PARSE_ERROR']
                              ['error_code']),
                        message=(validation_errors['DATETIME_PARSE_ERROR']
                                 ['error_mesg']),
                        field=key
                    )
                    found_errors.append(error_to_append)
                elif msg_error.startswith("unallowed value"):
                    error_to_append = dict(
                        code=(validation_errors['VALUE_OUT_OF_BOUNDS']
                              ['error_code']),
                        message=(validation_errors['VALUE_OUT_OF_BOUNDS']
                                 ['error_mesg'] +
                                 " : " + msg_error),
                        field=key
                    )
                    found_errors.append(error_to_append)

            elif isinstance(all_errors[key], dict):
                error_dict = all_errors[key]
                for entry in error_dict:
                    if re.match(
                        "must be of (string|integer|float|boolean|list) type",
                        error_dict[entry]
                    ):
                        error_to_append = dict(
                            code=(validation_errors
                                  ['INCORRECT_ELEMENT_TYPE_IN_DATASTRUCTURE']
                                  ['error_code']),
                            message=(
                                validation_errors
                                ['INCORRECT_ELEMENT_TYPE_IN_DATASTRUCTURE']
                                ['error_mesg'] +
                                " : " + error_dict[entry]),
                            field=key
                        )
                        found_errors.append(error_to_append)
                        break

        return found_errors


#
# Marshal fields
#
class ISODate(fields.Raw):

    """Format a datetime object in ISO format."""

    def format(self, dt):
        """Format a datetime object in ISO format."""
        try:
            if isinstance(dt, date):
                return six.text_type(dt.isoformat())
            else:
                return six.text_type(dt)
        except AttributeError as ae:
            raise fields.MarshallingException(ae)


class UTCISODateTime(fields.DateTime):

    """Format a datetime object in ISO format.

    Convert to UTC if necessary.
    """

    def format(self, dt):
        """Format a datetime object in ISO format.

        Convert to UTC if necessary.
        """
        try:
            if not dt.tzinfo:
                dt = dt.replace(tzinfo=tzlocal())
            return six.text_type(dt.astimezone(tzutc()).isoformat())
        except AttributeError as ae:
            raise fields.MarshallingException(ae)


class UTCISODateTimeString(fields.DateTime):

    """Format a string which represents a datetime in ISO format.

    Convert to UTC if necessary.
    """

    def format(self, value):
        """Format a string which represents a datetime in ISO format.

        Convert to UTC if necessary.
        """
        try:
            dt = parser.parse(value)
            if not dt.tzinfo:
                dt = dt.replace(tzinfo=tzlocal())
            return six.text_type(dt.astimezone(tzutc()).isoformat())
        except AttributeError as ae:
            raise fields.MarshallingException(ae)


#
# Decorators
#
def require_api_auth(*scopes):
    """Decorator to require API authentication using either API key or OAuth.

    Note, API key usage will be deprecated. Personal OAuth access tokens
    provide the same features as API keys.

    :param scopes: List of required OAuth scopes.
    """
    # Decorators specified in  ``method_decorators``  in Flask-Restful's
    # attribute is applied to a bound instance method, where as if you
    # decorate the class method is applied to an unbound instance. If you
    # are not accessing *args or **kwargs this doesn't matter. If you are
    # you can check if the method is bound using the following line:
    #   is_bound = hasattr(f, '__self__') and f.__self__

    def wrapper(f):
        """Wrap function with oauth require decorator."""
        from invenio_oauth2server.provider import oauth2
        f_oauth_required = oauth2.require_oauth()(f)

        @wraps(f)
        def decorated(*args, **kwargs):
            """OAuth 2.0 Authentication."""
            resp = f_oauth_required(*args, **kwargs)
            session.clear()
            return resp
        return decorated
    return wrapper


def require_oauth_scopes(*scopes):
    """Decorator to require a list of OAuth scopes.

    Decorator must be preceded by a ``require_api_auth()`` decorator.
    Note, API key authentication is bypassing this check
    """
    required_scopes = set(scopes)

    def wrapper(f):
        @wraps(f)
        def decorated(*args, **kwargs):
            # Variable requests.oauth is only defined for oauth requests (see
            # require_api_auth() above).
            if hasattr(request, 'oauth') and request.oauth is not None:
                token_scopes = set(request.oauth.access_token.scopes)
                if not required_scopes.issubset(token_scopes):
                    restful.abort(403)
            return f(*args, **kwargs)
        return decorated
    return wrapper


def require_header(header, value):
    """Decorator to test if proper content-type is provided."""
    def decorator(f):
        @wraps(f)
        def inner(*args, **kwargs):
            if header == 'Content-Type':
                test_value = request.headers.get(header, '').split(';')[0]
            else:
                test_value = request.headers.get(header, '')

            if (callable(value) and not value(test_value)) or \
                    test_value != value:
                msg = value if not callable(value) else value.__doc__
                restful.abort(
                    415,
                    message="Expected %s: %s" % (header, msg),
                    status=415,
                )
            return f(*args, **kwargs)
        return inner
    return decorator


def setup_app(app):
    """Setup api extension."""
    api = restful.Api(app=app)
    app.extensions['restful'] = api

    class RestfulRegistry(ModuleAutoDiscoveryRegistry):
        setup_func_name = 'setup_app'

        def register(self, module, *args, **kwargs):
            return super(RestfulRegistry, self).register(module, app, api,
                                                         *args, **kwargs)

    app.extensions['registry']['restful'] = RestfulRegistry(
        'restful', app=app, with_setup=True
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Test *Passlib* integration."""

from __future__ import absolute_import

from binascii import hexlify, unhexlify

from invenio_ext.passlib.hash import mysql_aes_decrypt, mysql_aes_encrypt
from invenio_testing import InvenioTestCase


class MySQLAESEncryptTestCase(InvenioTestCase):

    """Test of extension creation."""

    def test_mysql_aes_encrypt(self):
        """Test mysql_aes_encrypt."""
        self.assertEqual(
            hexlify(mysql_aes_encrypt("test", "key")),
            "9e9ce44cd9df2b201f51947e03bccbe2"
        )
        self.assertEqual(
            hexlify(mysql_aes_encrypt(u"test", "key")),
            "9e9ce44cd9df2b201f51947e03bccbe2"
        )
        self.assertEqual(
            hexlify(mysql_aes_encrypt("test", u"key")),
            "9e9ce44cd9df2b201f51947e03bccbe2"
        )
        self.assertEqual(
            hexlify(mysql_aes_encrypt(u"test", u"key")),
            "9e9ce44cd9df2b201f51947e03bccbe2"
        )
        self.assertRaises(AssertionError, mysql_aes_encrypt, object(), "key")
        self.assertRaises(AssertionError, mysql_aes_encrypt, "val", object())

    def test_mysql_aes_decrypt(self):
        """Test mysql_aes_encrypt."""
        self.assertEqual(
            mysql_aes_decrypt(unhexlify("9e9ce44cd9df2b201f51947e03bccbe2"),
                              "key"),
            "test"
        )
        self.assertEqual(
            mysql_aes_decrypt(unhexlify(u"9e9ce44cd9df2b201f51947e03bccbe2"),
                              u"key"),
            "test"
        )
        self.assertRaises(AssertionError, mysql_aes_decrypt, object(), "key")
        self.assertRaises(AssertionError, mysql_aes_decrypt, "val", object())


class PasslibTestCase(InvenioTestCase):

    """Test passlib extensions."""

    config = {
        'PASSLIB_SCHEMES': ['sha512_crypt', 'invenio_aes_encrypted_email'],
        'PASSLIB_DEPRECATED_SCHEMES': ['invenio_aes_encrypted_email']
    }

    def test_context(self):
        """Test passlib password context."""
        ctx = self.app.extensions['passlib']
        hashval = ctx.encrypt("test")
        assert hashval != "test"
        assert ctx.verify("test", hashval)
        assert not ctx.needs_update(hashval)
        assert ctx.encrypt("test") != ctx.encrypt("test")

    def test_invenio_aes_encrypted_email(self):
        """Test legacy encryption."""
        ctx = self.app.extensions['passlib']
        hashval = ctx.encrypt(
            "mypassword",
            scheme="invenio_aes_encrypted_email",
            user="info@invenio-software.org",
        )
        assert ctx.verify("mypassword", hashval,
                          scheme="invenio_aes_encrypted_email",
                          user="info@invenio-software.org", )
        assert ctx.needs_update(hashval)

    def test_unicode_regression(self):
        """Test legacy encryption."""
        ctx = self.app.extensions['passlib']
        hashval = ctx.encrypt(
            u"mypassword",
            scheme="invenio_aes_encrypted_email",
            user=u"info@invenio-software.org",
        )
        assert ctx.verify(u"mypassword", hashval,
                          scheme="invenio_aes_encrypted_email",
                          user=u"info@invenio-software.org", )
        assert ctx.needs_update(hashval)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Custom `Jinja2` extensions."""

import copy
import os

import six
from flask import _request_ctx_stack, current_app
from flask_assets import Environment, FlaskResolver
from jinja2 import nodes
from jinja2.ext import Extension
from webassets.bundle import is_url

from . import registry


class BundleExtension(Extension):

    """
    Jinja extension for css and js bundles.

    Definition of the required bundles.

    .. code-block:: jinja

        {%- bundles "jquery.js", "invenio.css" -%}
        {%- bundle "require.js" -%}

    Usage.

    .. code-block:: jinja

        {%- for bundle in get_bundle('js') %}
          <!-- {{ bundle.output }} -->
          {%- assets bundle %}
            <script type="text/javascript" src="{{ ASSET_URL }}"></script>
          {%- endassets %}
        {%- endfor %}
        </body>
        </html>
    """

    tags = set(('bundle', 'bundles'))

    @classmethod
    def storage(cls):
        """Store used bundles on request context stack."""
        ctx = _request_ctx_stack.top
        if ctx is not None:
            if not hasattr(ctx, "_bundles"):
                setattr(ctx, "_bundles", set())
            return ctx._bundles

    @classmethod
    def install(cls, app):
        """Install the extension into the application."""
        Environment.resolver_class = InvenioResolver
        env = Environment(app)
        env.url = "{0}/{1}/".format(app.static_url_path,
                                    app.config["ASSETS_BUNDLES_DIR"])
        env.directory = os.path.join(app.static_folder,
                                     app.config["ASSETS_BUNDLES_DIR"])
        env.append_path(app.static_folder)
        env.auto_build = app.config.get("ASSETS_AUTO_BUILD", True)

        # The filters less and requirejs don't have the same behaviour by
        # default. Make sure we are respecting that.
        app.config.setdefault("REQUIREJS_RUN_IN_DEBUG", False)
        # Fixing some paths as we forced the output directory with the
        # .directory
        app.config.setdefault("REQUIREJS_BASEURL", app.static_folder)
        requirejs_config = os.path.join(env.directory,
                                        app.config["REQUIREJS_CONFIG"])
        if not os.path.exists(requirejs_config):
            app.config["REQUIREJS_CONFIG"] = os.path.relpath(
                os.path.join(app.static_folder,
                             app.config["REQUIREJS_CONFIG"]),
                env.directory)

        app.jinja_env.add_extension(BundleExtension)
        app.context_processor(BundleExtension.inject)

    @classmethod
    def inject(cls):
        """Inject the get_bundle function into the jinja templates."""
        _bundles = {}

        def get_bundle(suffix):
            # lazy build the bundles
            if not _bundles:
                for pkg, bundle in registry.bundles:
                    if bundle.output in _bundles:
                        raise ValueError("{0} was already defined!"
                                         .format(bundle.output))
                    _bundles[bundle.output] = bundle

            env = current_app.jinja_env.assets_environment

            requirejs_debug = env.debug and \
                not current_app.config.get("REQUIREJS_RUN_IN_DEBUG")

            static_url_path = current_app.static_url_path + "/"
            bundles = []
            for bundle_name in cls.storage():
                if bundle_name.endswith(suffix):
                    bundle = _bundles[bundle_name]
                    if suffix == "css":
                        bundle.extra.update(rel="stylesheet")
                    bundles.append((bundle.weight, bundle))

            from webassets.filter import option

            def option__deepcopy__(value, memo):
                """Custom deepcopy implementation for ``option`` class."""
                return option(copy.deepcopy(value[0]),
                              copy.deepcopy(value[1]),
                              copy.deepcopy(value[2]))

            option.__deepcopy__ = option__deepcopy__

            for _, bundle in sorted(bundles):
                # A little bit of madness to read the "/" at the
                # beginning of the assets in ran in debug mode as well as
                # killing the filters if they are not wanted in debug mode.
                if env.debug:
                    # Create a deep copy to avoid filter removal from
                    # being cached
                    bundle_copy = copy.deepcopy(bundle)
                    bundle_copy.extra.update(static_url_path=static_url_path)
                    if bundle.has_filter("less"):
                        bundle_copy.extra.update(static_url_path="")
                    if bundle.has_filter("requirejs"):
                        if requirejs_debug:
                            bundle_copy.filters = None
                        else:
                            bundle_copy.extra.update(static_url_path="")
                    yield bundle_copy
                else:
                    yield bundle

        return dict(get_bundle=get_bundle)

    def __init__(self, environment):
        """Initialize the extension."""
        super(BundleExtension, self).__init__(environment)

    def _update(self, filename, bundles, caller):
        """Update the environment bundles.

        :return: empty html or html comment in debug mode.
        :rtype: str
        """
        self.storage().update(bundles)
        if current_app.debug:
            return "<!-- {0}: {1} -->\n".format(filename, ", ".join(bundles))
        else:
            return ''

    def parse(self, parser):
        """Parse the bundles block and feed the bundles environment.

        Bundles entries are replaced by an empty string.
        """
        lineno = next(parser.stream).lineno

        bundles = []
        while parser.stream.current.type != "block_end":
            value = parser.parse_expression()
            bundles.append(value)
            parser.stream.skip_if("comma")

        call = self.call_method("_update", args=[nodes.Const(parser.name),
                                                 nodes.List(bundles)])
        call_block = nodes.CallBlock(call, [], [], '')
        call_block.set_lineno(lineno)
        return call_block


class InvenioResolver(FlaskResolver):

    """Custom resource resolver for webassets."""

    def resolve_source(self, ctx, item):
        """Return the absolute path of the resource."""
        if not isinstance(item, six.string_types) or is_url(item):
            return item
        if item.startswith(ctx.url):
            item = item[len(ctx.url):]
        return self.search_for_source(ctx, item)

    def resolve_source_to_url(self, ctx, filepath, item):
        """Return the url of the resource.

        Displaying them as is in debug mode as the web server knows where to
        search for them.

        :py:meth:`webassets.env.Resolver.resolve_source_to_url`
        """
        if ctx.debug:
            return item
        return super(InvenioResolver, self).resolve_source_to_url(ctx,
                                                                  filepath,
                                                                  item)

    def search_for_source(self, ctx, item):
        """Return absolute path of the resource.

        :py:meth:`webassets.env.Resolver.search_for_source`

        :param ctx: environment
        :param item: resource filename
        :return: absolute path
        """
        try:
            if ctx.load_path:
                abspath = super(InvenioResolver, self) \
                    .search_load_path(ctx, item)
            else:
                abspath = super(InvenioResolver, self) \
                    .search_env_directory(ctx, item)
        except Exception:  # FIXME do not catch all!
            # If a file is missing in production (non-debug mode), we want
            # to not break and will use /dev/null instead. The exception
            # is caught and logged.
            if not current_app.debug:
                error = "Error loading asset file: {0}".format(item)
                current_app.logger.exception(error)
                abspath = "/dev/null"
            else:
                raise

        return abspath



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Implement various custom column types."""

from .guid import GUID
from .legacybiginteger import LegacyBigInteger
from .legacyboolean import LegacyBoolean
from .legacyinteger import LegacyInteger
from .legacymediuminteger import LegacyMediumInteger
from .legacysmallinteger import LegacySmallInteger
from .legacytinyinteger import LegacyTinyInteger
from .marshal_binary import MarshalBinary
from .pickle_binary import PickleBinary

__all__ = ('GUID', 'MarshalBinary', 'PickleBinary',
           'LegacyInteger', 'LegacyMediumInteger', 'LegacySmallInteger',
           'LegacyTinyInteger', 'LegacyBigInteger', 'LegacyBoolean')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
    invenio_ext.gravatar
    --------------------

    This module provides initialization and configuration for
    `flask_gravatar` module.
"""


def setup_app(app):
    """Initialize Gravatar extension."""
    from flask_gravatar import Gravatar
    gravatar = Gravatar(app,
                        size=app.config.get('GRAVATAR_SIZE', 100),
                        rating=app.config.get('GRAVATAR_RATING', 'g'),
                        default=app.config.get('GRAVATAR_DEFAULT', 'retro'),
                        force_default=False,
                        force_lower=False)
    del gravatar
    return app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Extension for password contexts via Passlib.

Support validation of multiple hashing algorithms to allow for easy migration
from one algorithm to another.

Example:

.. code-block:: console

   from invenio_ext.passlib import password_context
   hash = password_context.encrypt("mypassword")
   password_context.verify("mypassword", hash)
   password_context.needs_update(hash)

Invenio legacy support:

.. code-block:: console

   from invenio_ext.passlib import password_context
   hash = password_context.encrypt(
       "mypassword",
       scheme="invenio_aes_encrypted_email",
       user="info@invenio-software.org",
    )
   password_context.verify(
       "mypassword", hash
       scheme="invenio_aes_encrypted_email",
       user="info@invenio-software.org",
   )
   password_context.needs_update(hash)


Configuration Settings
----------------------
Invenio's default password hashing algorithms can be modified using the
following application settings:

============================ ==================================================
`PASSLIB_SCHEMES`            List of supported password hashing schemes. The
                             default password hashing scheme is the first item
                             in the list.
                             **Default:**
                             `['sha512_crypt', 'invenio_aes_encrypted_email']`
`PASSLIB_DEPRECATED_SCHEMES` List of password hashing schemes that are
                             deprecated (which results in the users' hash being
                             automatically upgrade on next login). Note, all
                             deprecated scheme must also be present in
                             `PASSLIB_SCHEMES`.
                             **Default:** `['invenio_aes_encrypted_email']`
============================ ==================================================
"""

from flask import current_app

from passlib.context import CryptContext
from passlib.registry import register_crypt_handler

from werkzeug.local import LocalProxy

from .hash import invenio_aes_encrypted_email

password_context = LocalProxy(
    lambda: current_app.extensions['passlib']
)


class Passlib(object):

    """Flask-Passlib integration."""

    def __init__(self, app=None):
        """Initialize extension."""
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Initialize application."""
        # Register Invenio legacy password hashing.
        register_crypt_handler(invenio_aes_encrypted_email)

        app.config.setdefault(
            'PASSLIB_SCHEMES',
            ['sha512_crypt', 'invenio_aes_encrypted_email']
        )
        app.config.setdefault(
            'PASSLIB_DEPRECATED_SCHEMES',
            ['invenio_aes_encrypted_email']
        )

        # Create password context.
        app.extensions['passlib'] = CryptContext(
            schemes=app.config['PASSLIB_SCHEMES'],
            deprecated=app.config['PASSLIB_DEPRECATED_SCHEMES']
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Filter for assets."""

from __future__ import unicode_literals

from flask import current_app
from webassets.filter.cssrewrite.base import PatternRewriter, urltag_re

__all__ = (
    'CSSUrlFixer',
)


# Helper class as a workaround for the strange cleancss URL handling
class CSSUrlFixer(PatternRewriter):

    """Helper to fix `url(...)` entries in CSS files."""

    patterns = {
        'rewrite_url': urltag_re
    }

    def __init__(self, subpath):
        """Create new CSSUrlFixer.

        :param subpath: relative subpath to the site URL, where all resources
            are located.
        """
        self.subpath = subpath
        super(CSSUrlFixer, self).__init__()

    def rewrite_url(self, matches):
        """Rewrite found URL pattern."""
        # Get the regex matches; note how we maintain the exact
        # whitespace around the actual url; we'll indeed only
        # replace the url itself.
        text_before = matches.groups()[0]
        url = matches.groups()[1]
        text_after = matches.groups()[2]

        # Normalize the url: remove quotes
        quotes_used = ''
        if url[:1] in '"\'':
            quotes_used = url[:1]
            url = url[1:]
        if url[-1:] in '"\'':
            url = url[:-1]

        url = self.replace_url(url) or url

        result = 'url({before}{quotes}{url}{quotes}{after})'.format(
            before=text_before, quotes=quotes_used, url=url, after=text_after
        )
        return result

    def replace_url(self, url):
        """Replace given URL with a new one."""
        return "{}/{}/{}".format(
            current_app.config.get('CFG_SITE_URL'),
            self.subpath,
            url
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Invenio mail sending utilities.

send_email() is the main API function people should be using; just check out
its docstring.
"""

from __future__ import absolute_import

import os
import re
import sys
import urllib2
from email import Encoders
from email.Header import Header
from email.MIMEBase import MIMEBase
from email.MIMEImage import MIMEImage
from email.MIMEMultipart import MIMEMultipart
from email.Utils import formatdate
from formatter import AbstractFormatter, DumbWriter
from mimetypes import MimeTypes
from time import sleep

from flask_email.message import EmailMessage, EmailMultiAlternatives
from invenio_base.globals import cfg
from invenio_base.helpers import unicodifier
from invenio_base.i18n import _
from six import StringIO, iteritems

from .errors import EmailError


def default_ln(ln):
    """Default ln."""
    cfg.get('CFG_SITE_LANG') if ln is None else ln


def getContentType(pageUrl):
    """Get content type."""
    page = urllib2.urlopen(pageUrl)
    pageHeaders = page.headers
    contentType = pageHeaders.getheader('content-type')
    return contentType

def setup_app(app):
    """
    Prepare application config from Invenio configuration.

    @see: https://flask-email.readthedocs.org/en/latest/#configuration
    """
    cfg = app.config

    app.config.setdefault('EMAIL_BACKEND', cfg.get(
        'CFG_EMAIL_BACKEND', 'flask_email.backends.smtp.Mail'))
    app.config.setdefault('DEFAULT_FROM_EMAIL', cfg['CFG_SITE_SUPPORT_EMAIL'])
    app.config.setdefault('SERVER_EMAIL', cfg['CFG_SITE_ADMIN_EMAIL'])
    app.config.setdefault('ADMINS', (('', cfg['CFG_SITE_ADMIN_EMAIL']),))
    app.config.setdefault('MANAGERS', (cfg['CFG_SITE_SUPPORT_EMAIL'], ))

    CFG_MISCUTIL_SMTP_HOST = cfg.get('CFG_MISCUTIL_SMTP_HOST')
    CFG_MISCUTIL_SMTP_PORT = cfg.get('CFG_MISCUTIL_SMTP_PORT')
    CFG_MISCUTIL_SMTP_USER = cfg.get('CFG_MISCUTIL_SMTP_USER', '')
    CFG_MISCUTIL_SMTP_PASS = cfg.get('CFG_MISCUTIL_SMTP_PASS', '')
    CFG_MISCUTIL_SMTP_TLS = cfg.get('CFG_MISCUTIL_SMTP_TLS', False)

    app.config.setdefault('EMAIL_HOST', CFG_MISCUTIL_SMTP_HOST)
    app.config.setdefault('EMAIL_PORT', CFG_MISCUTIL_SMTP_PORT)
    app.config.setdefault('EMAIL_HOST_USER', CFG_MISCUTIL_SMTP_USER)
    app.config.setdefault('EMAIL_HOST_PASSWORD', CFG_MISCUTIL_SMTP_PASS)
    app.config.setdefault('EMAIL_USE_TLS', CFG_MISCUTIL_SMTP_TLS)
    # app.config['EMAIL_USE_SSL']: defaults to False

    app.config.setdefault('EMAIL_FILE_PATH', cfg['CFG_LOGDIR'])
    return app


def scheduled_send_email(fromaddr,
                         toaddr,
                         subject="",
                         content="",
                         header=None,
                         footer=None,
                         copy_to_admin=0,
                         attempt_times=1,
                         attempt_sleeptime=10,
                         user=None,
                         other_bibtasklet_arguments=None,
                         replytoaddr="",
                         bccaddr="",
                         ):
    """
    Like send_email, but send an email via the bibsched
    infrastructure.
    @param fromaddr: sender
    @type fromaddr: string
    @param toaddr: list of receivers
    @type toaddr: string (comma separated) or list of strings
    @param subject: the subject
    @param content: the body of the message
    @param header: optional header, otherwise default is used
    @param footer: optional footer, otherwise default is used
    @param copy_to_admin: set to 1 in order to send email the admins
    @param attempt_times: try at least n times before giving up sending
    @param attempt_sleeptime: number of seconds to sleep between two attempts
    @param user: the user name to user when scheduling the bibtasklet. If
        None, the sender will be used
    @param other_bibtasklet_arguments: other arguments to append to the list
        of arguments to the call of task_low_level_submission
    @param replytoaddr: [string or list-of-strings] to be used for the
                        reply-to header of the email (if string, then
                        receivers are separated by ',')
    @param bccaddr: [string or list-of-strings] to be used for BCC header
                     of the email
                    (if string, then receivers are separated by ',')
    @return: the scheduled bibtasklet
    """
    raise RuntimeError('scheduled_send_email has been removed.')


def send_email(fromaddr,
               toaddr,
               subject="",
               content="",
               html_content='',
               html_images=None,
               header=None,
               footer=None,
               html_header=None,
               html_footer=None,
               copy_to_admin=0,
               attempt_times=1,
               attempt_sleeptime=10,
               debug_level=0,
               ln=None,
               charset=None,
               replytoaddr="",
               attachments=None,
               bccaddr="",
               forward_failures_to_admin=True,
               ):
    """Send a forged email to TOADDR from FROMADDR with message created from subjet, content and possibly
    header and footer.
    @param fromaddr: [string] sender
    @param toaddr: [string or list-of-strings] list of receivers (if string, then
                   receivers are separated by ','). BEWARE: If more than once receiptiant is given,
                   the receivers are put in BCC and To will be "Undisclosed.Recipients:".
    @param subject: [string] subject of the email
    @param content: [string] content of the email
    @param html_content: [string] html version of the email
    @param html_images: [dict] dictionary of image id, image path
    @param header: [string] header to add, None for the Default
    @param footer: [string] footer to add, None for the Default
    @param html_header: [string] header to add to the html part, None for the Default
    @param html_footer: [string] footer to add to the html part, None for the Default
    @param copy_to_admin: [int] if 1 add CFG_SITE_ADMIN_EMAIL in receivers
    @param attempt_times: [int] number of tries
    @param attempt_sleeptime: [int] seconds in between tries
    @param debug_level: [int] debug level
    @param ln: [string] invenio language
    @param charset: [string] the content charset. By default is None which means
    to try to encode the email as ascii, then latin1 then utf-8.
    @param replytoaddr: [string or list-of-strings] to be used for the
                        reply-to header of the email (if string, then
                        receivers are separated by ',')
    @param attachments: list of paths of files to be attached. Alternatively,
        every element of the list could be a tuple: (filename, mimetype)
    @param bccaddr: [string or list-of-strings] to be used for BCC header of the email
                    (if string, then receivers are separated by ',')
    @param forward_failures_to_admin: [bool] prevents infinite recursion
                                             in case of admin reporting,
                                             when the problem is not in
                                             the e-mail address format,
                                             but rather in the network

    If sending fails, try to send it ATTEMPT_TIMES, and wait for
    ATTEMPT_SLEEPTIME seconds in between tries.

    e.g.:
    send_email('foo.bar@cern.ch', 'bar.foo@cern.ch', 'Let\'s try!'', 'check 1234', '<strong>check</strong> <em>1234</em><img src="cid:image1">', {'image1': '/tmp/quantum.jpg'})

    @return: [bool]: True if email was sent okay, False if it was not.
    """
    from invenio_ext.logging import register_exception
    ln = default_ln(ln)

    if html_images is None:
        html_images = {}

    if not isinstance(toaddr, list):
        toaddr = toaddr.strip().split(',')
    toaddr = remove_temporary_emails(toaddr)

    # More than one address, let's use Bcc in place of To
    usebcc = len(toaddr) > 1

    if copy_to_admin:
        if cfg['CFG_SITE_ADMIN_EMAIL'] not in toaddr:
            toaddr.append(cfg['CFG_SITE_ADMIN_EMAIL'])

    if not isinstance(bccaddr, list):
        bccaddr = bccaddr.strip().split(',')

    msg = forge_email(fromaddr, toaddr, subject, content, html_content,
                      html_images, usebcc, header, footer, html_header,
                      html_footer, ln, charset, replytoaddr, attachments,
                      bccaddr)

    if attempt_times < 1 or not toaddr:
        try:
            raise EmailError(_(
                'The system is not attempting to send an email from %(x_from)s'
                ', to %(x_to)s, with body %(x_body)s.', x_from=fromaddr,
                x_to=toaddr, x_body=content))
        except EmailError:
            register_exception()
        return False
    sent = False
    failure_reason = ''
    while not sent and attempt_times > 0:
        try:
            sent = msg.send()
        except Exception as e:
            failure_reason = str(e)
            register_exception()
            if debug_level > 1:
                try:
                    raise EmailError(_('Error in sending message. \
                        Waiting %(sec)s seconds. Exception is %(exc)s, \
                        while sending email from %(sender)s to %(receipient)s \
                        with body %(email_body)s.',
                                       sec=attempt_sleeptime,
                                       exc=sys.exc_info()[0],
                                       sender=fromaddr,
                                       receipient=toaddr,
                                       email_body=content))
                except EmailError:
                    register_exception()
        if not sent:
            attempt_times -= 1
            if attempt_times > 0:  # sleep only if we shall retry again
                sleep(attempt_sleeptime)
    if not sent:
        # report failure to the admin with the intended message, its
        # sender and recipients
        if forward_failures_to_admin:
            # prepend '> ' to every line of the original message
            quoted_body = '> ' + '> '.join(content.splitlines(True))

            # define and fill in the report template
            admin_report_subject = _(
                'Error while sending an email: %(x_subject)s',
                x_subject=subject)
            admin_report_body = _(
                "\nError while sending an email.\n"
                "Reason: %(x_reason)s\n"
                "Sender: \"%(x_sender)s\"\n"
                "Recipient(s): \"%(x_recipient)s\"\n\n"
                "The content of the mail was as follows:\n"
                "%(x_body)s",
                x_reason=failure_reason,
                x_sender=fromaddr,
                x_recipient=', '.join(toaddr),
                x_body=quoted_body)

            send_email(
                cfg['CFG_SITE_ADMIN_EMAIL'],
                cfg['CFG_SITE_ADMIN_EMAIL'],
                admin_report_subject,
                admin_report_body,
                forward_failures_to_admin=False)

        try:
            raise EmailError(_(
                'Error in sending email from %(x_from)s to %(x_to)s with body'
                '%(x_body)s.', x_from=fromaddr, x_to=toaddr, x_body=content))
        except EmailError:
            register_exception()
    return sent


def attach_embed_image(email, image_id, image_path):
    """
    Attach an image to the email.
    """
    with open(image_path, 'rb') as image_data:
        img = MIMEImage(image_data.read())
        img.add_header('Content-ID', '<%s>' % image_id)
        img.add_header(
            'Content-Disposition',
            'attachment',
            filename=os.path.split(image_path)[1])
        email.attach(img)


def forge_email(fromaddr, toaddr, subject, content, html_content='',
                html_images=None, usebcc=False, header=None, footer=None,
                html_header=None, html_footer=None, ln=None,
                charset=None, replytoaddr="", attachments=None, bccaddr=""):
    """Prepare email. Add header and footer if needed.
    @param fromaddr: [string] sender
    @param toaddr: [string or list-of-strings] list of receivers (if string, then
                   receivers are separated by ',')
    @param usebcc: [bool] True for using Bcc in place of To
    @param subject: [string] subject of the email
    @param content: [string] content of the email
    @param html_content: [string] html version of the email
    @param html_images: [dict] dictionary of image id, image path
    @param header: [string] None for the default header
    @param footer: [string] None for the default footer
    @param ln: language
    @charset: [string] the content charset. By default is None which means
    to try to encode the email as ascii, then latin1 then utf-8.
    @param replytoaddr: [string or list-of-strings] to be used for the
                        reply-to header of the email (if string, then
                        receivers are separated by ',')
    @param attachments: list of paths of files to be attached. Alternatively,
        every element of the list could be a tuple: (filename, mimetype)
    @param bccaddr: [string or list-of-strings] to be used for BCC header of the email
                    (if string, then receivers are separated by ',')
    @return: forged email as an EmailMessage object"""
    from invenio_ext.template import render_template_to_string

    ln = default_ln(ln)
    if html_images is None:
        html_images = {}

    content = render_template_to_string('mail_text.tpl',
                                        content=unicodifier(content),
                                        header=unicodifier(header),
                                        footer=unicodifier(footer)
                                        ).encode('utf8')

    if isinstance(toaddr, list):
        toaddr = ','.join(toaddr)

    if isinstance(bccaddr, list):
        bccaddr = ','.join(bccaddr)

    if isinstance(replytoaddr, list):
        replytoaddr = ','.join(replytoaddr)

    toaddr = remove_temporary_emails(toaddr)

    headers = {}
    kwargs = {'to': [], 'cc': [], 'bcc': []}

    if replytoaddr:
        headers['Reply-To'] = replytoaddr
    if usebcc:
        headers['Bcc'] = bccaddr
        kwargs['bcc'] = toaddr.split(',') + bccaddr.split(',')
        kwargs['to'] = ['Undisclosed.Recipients:']
    else:
        kwargs['to'] = toaddr.split(',')
    headers['From'] = fromaddr
    headers['Date'] = formatdate(localtime=True)
    headers['User-Agent'] = 'Invenio %s at %s' % (cfg['CFG_VERSION'],
                                                  cfg['CFG_SITE_URL'])

    if html_content:
        html_content = render_template_to_string(
            'mail_html.tpl',
            content=unicodifier(html_content),
            header=unicodifier(html_header),
            footer=unicodifier(html_footer)
        ).encode('utf8')

        msg_root = EmailMultiAlternatives(subject=subject, body=content,
                                          from_email=fromaddr,
                                          headers=headers, **kwargs)
        msg_root.attach_alternative(html_content, "text/html")

        # if not html_images:
        #    # No image? Attach the HTML to the root
        #    msg_root.attach(msg_text)
        # else:
        if html_images:
            # Image(s)? Attach the HTML and image(s) as children of a
            # "related" block
            msg_related = MIMEMultipart('related')
            # msg_related.attach(msg_text)
            for image_id, image_path in iteritems(html_images):
                attach_embed_image(msg_related, image_id, image_path)
            msg_root.attach(msg_related)
    else:
        msg_root = EmailMessage(subject=subject, body=content,
                                from_email=fromaddr, headers=headers, **kwargs)

    if attachments:
        _mimes = MimeTypes(strict=False)
        for attachment in attachments:
            try:
                mime = None
                if type(attachment) in (list, tuple):
                    attachment, mime = attachment
                if mime is None:
                    # Automatic guessing of mimetype
                    mime = _mimes.guess_type(attachment)[0]
                if mime is None:
                    ext = _mimes.guess_extension(getContentType(attachment))
                    mime = _mimes.guess_type("foo" + ext)[0]
                if not mime:
                    mime = 'application/octet-stream'
                part = MIMEBase(*mime.split('/', 1))
                part.set_payload(open(attachment, 'rb').read())
                Encoders.encode_base64(part)
                part.add_header(
                    'Content-Disposition',
                    'attachment; filename="%s"' %
                    os.path.basename(attachment))
                msg_root.attach(part)
            except:
                from invenio_ext.logging import register_exception
                register_exception(
                    alert_admin=True,
                    prefix="Can't attach %s" %
                    attachment)

    return msg_root

RE_NEWLINES = re.compile(r'<br\s*/?>|</p>', re.I)
RE_SPACES = re.compile(r'\s+')
RE_HTML_TAGS = re.compile(r'<.+?>')


def email_strip_html(html_content):
    """Strip html tags from html_content, trying to respect formatting."""
    html_content = RE_SPACES.sub(' ', html_content)
    html_content = RE_NEWLINES.sub('\n', html_content)
    html_content = RE_HTML_TAGS.sub('', html_content)
    html_content = html_content.split('\n')
    out = StringIO()
    out_format = AbstractFormatter(DumbWriter(out))
    for row in html_content:
        out_format.add_flowing_data(row)
        out_format.end_paragraph(1)
    return out.getvalue()


def remove_temporary_emails(emails):
    """
    Removes the temporary emails (which are constructed randomly when user logs in
    with an external authentication provider which doesn't supply an email
    address) from an email list.

    @param emails: email list (if string, then receivers are separated by ',')
    @type emails: [str]|str

    @rtype: list|str
    """
    from invenio_access.local_config import CFG_TEMP_EMAIL_ADDRESS
    _RE_TEMPORARY_EMAIL = re.compile(CFG_TEMP_EMAIL_ADDRESS % r'.+?', re.I)

    if type(emails) in (str, unicode):
        emails = [email.strip()
                  for email in emails.split(',') if email.strip()]
        emails = [
            email for email in emails if not _RE_TEMPORARY_EMAIL.match(email)]
        return ','.join(emails)
    else:
        return [
            email for email in emails if not _RE_TEMPORARY_EMAIL.match(email)]


def get_mail_header(value):
    """
    Return a MIME-compliant header-string. Will join lists of strings
    into one string with comma (,) as separator.
    """
    if not isinstance(value, basestring):
        value = ','.join(value)
    try:
        value = value.encode('ascii')
    except (UnicodeEncodeError, UnicodeDecodeError):
        value = Header(value, 'utf-8')
    return value



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Initialization and configuration for `flask_sqlalchemy`."""


from flask_registry import ModuleAutoDiscoveryRegistry, RegistryProxy

from flask_sqlalchemy import SQLAlchemy as FlaskSQLAlchemy

from invenio_ext.sqlalchemy.types import LegacyBigInteger, LegacyBoolean, \
    LegacyInteger, LegacyMediumInteger, LegacySmallInteger, LegacyTinyInteger

import sqlalchemy

from sqlalchemy import event, types as engine_types
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.pool import Pool

from sqlalchemy_utils import JSONType

from .expressions import AsBINARY
from .types import GUID, MarshalBinary, PickleBinary
from .utils import get_model_type


def _include_sqlalchemy(obj, engine=None):
    """Init all required SQLAlchemy's types."""
    # for module in sqlalchemy, sqlalchemy.orm:
    #    for key in module.__all__:
    #        if not hasattr(obj, key):
    #            setattr(obj, key,
    #                    getattr(module, key))

    if engine == 'mysql':
        from sqlalchemy.dialects import mysql as engine_types
    else:
        from sqlalchemy import types as engine_types

    # Length is provided to JSONType to ensure MySQL uses LONGTEXT instead
    # of TEXT which only provides for 64kb storage compared to 4gb for
    # LONGTEXT.
    setattr(obj, 'JSON', JSONType(length=2 ** 32 - 2))
    setattr(obj, 'Char', engine_types.CHAR)
    try:
        setattr(obj, 'TinyText', engine_types.TINYTEXT)
    except AttributeError:
        setattr(obj, 'TinyText', engine_types.TEXT)
    setattr(obj, 'hybrid_property', hybrid_property)
    try:
        setattr(obj, 'Double', engine_types.DOUBLE)
    except AttributeError:
        setattr(obj, 'Double', engine_types.FLOAT)
    setattr(obj, 'Binary', sqlalchemy.types.LargeBinary)
    setattr(obj, 'iBinary', sqlalchemy.types.LargeBinary)
    setattr(obj, 'iLargeBinary', sqlalchemy.types.LargeBinary)
    setattr(obj, 'iMediumBinary', sqlalchemy.types.LargeBinary)
    setattr(obj, 'UUID', GUID)
    setattr(obj, 'Integer', LegacyInteger)
    setattr(obj, 'MediumInteger', LegacyMediumInteger)
    setattr(obj, 'SmallInteger', LegacySmallInteger)
    setattr(obj, 'TinyInteger', LegacyTinyInteger)
    setattr(obj, 'BigInteger', LegacyBigInteger)
    setattr(obj, 'Boolean', LegacyBoolean)

    if engine == 'mysql':
        from .engines import mysql as dummy_mysql  # noqa
    #    module = invenio.sqlalchemyutils_mysql
    #    for key in module.__dict__:
    #        setattr(obj, key,
    #                getattr(module, key))

    obj.AsBINARY = AsBINARY
    obj.MarshalBinary = MarshalBinary
    obj.PickleBinary = PickleBinary

    # Overwrite :meth:`MutableDick.update` to detect changes.
    from sqlalchemy.ext.mutable import MutableDict

    def update_mutable_dict(self, *args, **kwargs):
        super(MutableDict, self).update(*args, **kwargs)
        self.changed()

    MutableDict.update = update_mutable_dict
    obj.MutableDict = MutableDict


# TODO check if we can comment this without problems
# @compiles(types.Text, 'postgresql')
# @compiles(sqlalchemy.dialects.postgresql.TEXT, 'postgresql')
# def compile_text(element, compiler, **kw):
#     """Redefine Text filed type for PostgreSQL."""
#     return 'TEXT'


@compiles(engine_types.VARBINARY, 'postgresql')
def compile_text(element, compiler, **kw):
    """Redefine VARBINARY filed type for PostgreSQL."""
    return 'BYTEA'


def autocommit_on_checkin(dbapi_con, con_record):
    """Call autocommit on raw mysql connection for fixing bug in MySQL 5.5."""
    try:
        dbapi_con.autocommit(True)
    except Exception:
        pass
        # FIXME
        # from invenio_ext.logging import register_exception
        # register_exception()

# Possibly register globally.
# event.listen(Pool, 'checkin', autocommit_on_checkin)


class SQLAlchemy(FlaskSQLAlchemy):

    """Database object."""

    def init_app(self, app):
        """Init application."""
        super(self.__class__, self).init_app(app)
        engine = app.config.get('CFG_DATABASE_TYPE', 'mysql')
        self.Model = get_model_type(self.Model)
        if engine == 'mysql':
            # Override MySQL parameters to force InnoDB engine
            mysql_parameters = {'keep_existing': True,
                                'extend_existing': False,
                                'mysql_engine': 'InnoDB',
                                'mysql_charset': 'utf8',
                                'sql_mode': 'ansi_quotes'}

            original_table = self.Table

            def table_with_innodb(*args, **kwargs):
                """Use same MySQL parameters that are used for ORM models."""
                new_kwargs = dict(mysql_parameters)
                new_kwargs.update(kwargs)
                return original_table(*args, **new_kwargs)

            self.Table = table_with_innodb
            self.Model.__table_args__ = mysql_parameters

        _include_sqlalchemy(self, engine=engine)

    def __getattr__(self, name):
        """
        Called when the normal mechanism fails.

        This is only called when the normal mechanism fails,
        so in practice should never be called.
        It is only provided to satisfy pylint that it is okay not to
        raise E1101 errors in the client code.

        :see http://stackoverflow.com/a/3515234/780928
        """
        raise AttributeError("%r instance has no attribute %r" % (self, name))

    def schemadiff(self, excludeTables=None):
        """Generate a schema diff."""
        from migrate.versioning import schemadiff
        return schemadiff \
            .getDiffOfModelAgainstDatabase(self.metadata,
                                           self.engine,
                                           excludeTables=excludeTables)

    def apply_driver_hacks(self, app, info, options):
        """Called before engine creation."""
        # Don't forget to apply hacks defined on parent object.
        super(self.__class__, self).apply_driver_hacks(app, info, options)
        if info.drivername == 'mysql':
            options.setdefault('execution_options', {
                # Autocommit cause Exception in SQLAlchemy >= 0.9.
                # @see http://docs.sqlalchemy.org/en/rel_0_9/
                #   core/connections.html#understanding-autocommit
                # 'autocommit': True,
                'use_unicode': False,
                'charset': 'utf8mb4',
            })
            event.listen(Pool, 'checkin', autocommit_on_checkin)
        elif info.drivername == 'postgresql':
            from psycopg2.extensions import adapt, register_adapter
            from werkzeug.local import LocalProxy

            def adapt_proxy(proxy):
                return adapt(proxy._get_current_object())

            register_adapter(LocalProxy, adapt_proxy)


db = SQLAlchemy()
"""
    Provides access to :class:`~.SQLAlchemy` instance.
"""

models = RegistryProxy('models', ModuleAutoDiscoveryRegistry, 'models')


def setup_app(app):
    """Setup SQLAlchemy extension."""
    if 'SQLALCHEMY_DATABASE_URI' not in app.config:
        from sqlalchemy.engine.url import URL
        cfg = app.config

        app.config['SQLALCHEMY_DATABASE_URI'] = URL(
            cfg.get('CFG_DATABASE_TYPE', 'mysql'),
            username=cfg.get('CFG_DATABASE_USER'),
            password=cfg.get('CFG_DATABASE_PASS'),
            host=cfg.get('CFG_DATABASE_HOST'),
            database=cfg.get('CFG_DATABASE_NAME'),
            port=cfg.get('CFG_DATABASE_PORT'),
        )

    # Let's initialize database.
    db.init_app(app)

    return app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2012, 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Additional extensions and filters for jinja2 module."""

import re
import os

from datetime import datetime

from flask import _request_ctx_stack, g, request, url_for

from flask_login import current_user

from jinja2 import ChoiceLoader

from six import iteritems, string_types

from werkzeug.routing import BuildError

from .bccache import BytecodeCacheWithConfig
from .context_processor import setup_app as context_processor_setup_app
from .loader import OrderAwareDispatchingJinjaLoader

ENV_PREFIX = '_collected_'


def render_template_to_string(input, _from_string=False, **context):
    """Render a template from the template folder with the given context.

    Code based on
    `<https://github.com/mitsuhiko/flask/blob/master/flask/templating.py>`_

    :param input: the string template, or name of the template to be
                  rendered, or an iterable with template names
                  the first one existing will be rendered
    :param context: the variables that should be available in the
                    context of the template.
    :return: a string

    """
    ctx = _request_ctx_stack.top
    ctx.app.update_template_context(context)
    if _from_string:
        template = ctx.app.jinja_env.from_string(input)
    else:
        template = ctx.app.jinja_env.get_or_select_template(input)
    return template.render(context)


def inject_utils():
    """Inject variables and functions to jinja execution context.

    In particular it will add:

    - ``url_for``: an Invenio specific wrapper of Flask url_for, that will let
      you obtain URLs for non Flask-native handlers (i.e. not yet ported
      Invenio URLs)
    - ``_``: this can be used to automatically translate a given string.
    - ``is_language_rtl``: True if the chosen language should be read right to
      left.
    """
    from invenio_base.i18n import is_language_rtl
    from invenio_utils.url import create_url, get_canonical_and_alternates_urls

    def invenio_url_for(endpoint, **values):
        try:
            return url_for(endpoint, **values)
        except BuildError:
            if re.match("https?://", endpoint, re.IGNORECASE):
                return endpoint
            if endpoint.startswith('.'):
                endpoint = request.blueprint + endpoint
            url = create_url('/' + endpoint.replace('.', '/'), values, False)
            return url.decode('utf-8')

    user = current_user._get_current_object()
    canonical_url, alternate_urls = get_canonical_and_alternates_urls(
        request.path)
    alternate_urls = dict((ln.replace('_', '-'), alternate_url)
                          for ln, alternate_url in iteritems(alternate_urls))
    return dict(
        current_user=user,
        is_language_rtl=is_language_rtl,
        canonical_url=canonical_url,
        alternate_urls=alternate_urls,
        url_for=invenio_url_for,
    )


def setup_app(app):
    """
    Extend application template filters with custom filters and fixes.

    List of applied filters:

    - filesizeformat
    - path_join
    - quoted_txt2html
    - invenio_format_date
    - invenio_pretty_date
    - invenio_url_args
    """
    from . import config
    app.config.from_object(config)

    context_processor_setup_app(app)
    app.context_processor(inject_utils)

    if app.config.get('JINJA2_BCCACHE', False):
        app.jinja_options = dict(
            app.jinja_options,
            auto_reload=app.config.get('JINJA2_BCCACHE_AUTO_RELOAD', False),
            cache_size=app.config.get('JINJA2_BCCACHE_SIZE', -1),
            bytecode_cache=BytecodeCacheWithConfig(app))

    # Let's customize the template loader to look into packages
    # and application templates folders.
    jinja_loader = ChoiceLoader([
        OrderAwareDispatchingJinjaLoader(app),
        app.jinja_loader,
    ])
    app.jinja_loader = jinja_loader

    for ext in app.config.get('JINJA2_EXTENSIONS', []):
        try:
            app.jinja_env.add_extension(ext)
        except Exception:
            app.logger.exception(
                'Problem with loading extension: "{0}"'.format(ext))

    def test_not_empty(v):
        return v is not None and v != ''

    @app.template_filter('u')
    def tounicode(value):
        if isinstance(value, str):
            return value.decode('utf8')
        return value

    @app.template_filter('s')
    def tostr(value):
        if not isinstance(value, str):
            if isinstance(value, unicode):
                value = value.encode('utf8')
            value = str(value)
        return value

    @app.template_filter('prefix')
    def _prefix(value, prefix=''):
        return prefix + value if test_not_empty(value) else ''

    @app.template_filter('suffix')
    def _suffix(value, suffix=''):
        return value + suffix if test_not_empty(value) else ''

    @app.template_filter('wrap')
    def _wrap(value, prefix='', suffix=''):
        return prefix + value + suffix if test_not_empty(value) else ''

    @app.template_filter('sentences')
    def _sentences(value, limit, separator='. '):
        """Return first `limit` number of sentences ending by `separator`."""
        return separator.join(value.split(separator)[:limit])

    @app.template_filter('path_join')
    def _os_path_join(d):
        """Shortcut for `os.path.join`."""
        return os.path.join(*d)

    @app.template_filter('quoted_txt2html')
    def _quoted_txt2html(*args, **kwargs):
        from invenio_utils.mail import email_quoted_txt2html
        return email_quoted_txt2html(*args, **kwargs)

    @app.template_filter('invenio_format_date')
    def _format_date(date):
        """
        Format a date into a human friendly format.

        It uses :py:func:`invenio.utils.date.convert_datetext_to_dategui`
        """
        from invenio_utils.date import convert_datetext_to_dategui
        if isinstance(date, datetime):
            return convert_datestruct_to_dategui(
                date.timetuple(),
                getattr(g, 'ln', app.config['CFG_SITE_LANG'])).decode('utf-8')
        return convert_datetext_to_dategui(
            date, getattr(g, 'ln', app.config['CFG_SITE_LANG'])
        ).decode('utf-8')

    @app.template_filter('invenio_pretty_date')
    def _pretty_date(date):
        """
        Format a timestamp into a human friendly format.

        It uses :py:func:`invenio.utils.date.pretty_date`
        """
        from invenio_utils.date import pretty_date
        if isinstance(date, datetime) or isinstance(date, string_types):
            return pretty_date(
                date, ln=getattr(g, 'ln', app.config['CFG_SITE_LANG']))
        return date

    @app.template_filter('invenio_url_args')
    def _url_args(d, append=u'?', filter=[]):
        from jinja2.utils import escape
        rv = append + u'&'.join(
            u'%s=%s' % (escape(key), escape(value))
            for key, value in d.iteritems(True)
            if value is not None and key not in filter
            # and not isinstance(value, Undefined)
        )
        return rv

    return app



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Initialize and configure Flask-Script extension.

Configuration
^^^^^^^^^^^^^
The following configuration variables are provided:

===================== =======================================================
`bind address`        Preferred binding address of the server. Can be used to
                      select a specific interface or to bind to all via
                      `0.0.0.0`.
`bind port`           Preferred binding port of the server. Can differ from
                      the one stated in `CFG_SITE_URL` so it can be accessed
                      via reverse proxy.
===================== =======================================================

They are assigned by the following parameters, in decreasing priority:

1. Command line arguments of `inveniomanage runserver`
2. `SERVER_BIND_ADDRESS` and `SERVER_BIND_PORT` configuration
3. Values guessed from `CFG_SITE_URL`
4. Defaults (`127.0.0.1:80`)
"""

from __future__ import print_function

import functools

import re

import ssl

from types import FunctionType

from flask import current_app, flash

from flask_registry import ModuleAutoDiscoveryRegistry, RegistryProxy

from flask_script import Manager as FlaskExtManager
from flask_script.commands import Clean, Server, Shell, ShowUrls

from invenio_base.signals import post_command, pre_command

from six.moves import urllib

from werkzeug.utils import find_modules, import_string


def change_command_name(method=None, new_name=None):
    """Change command name to `new_name` or replace '_' by '-'."""
    if method is None:
        return functools.partial(change_command_name, new_name=new_name)

    if new_name is None:
        new_name = method.__name__.replace('_', '-')
    method.__name__ = new_name

    return method


def generate_secret_key():
    """Generate secret key."""
    import string
    import random

    rng = random.SystemRandom()
    return ''.join(
        rng.choice(string.ascii_letters + string.digits)
        for dummy in range(0, 256)
    )


def print_progress(p, L=40, prefix='', suffix=''):
    """Print textual progress bar."""
    bricks = int(p * L)
    print('\r{prefix} [{bricks}{spaces}] {progress}% {suffix}'.format(
        prefix=prefix, suffix=suffix,
        bricks='#' * bricks, spaces=' ' * (L - bricks),
        progress=int(p * 100),
    ), end=' ')


def check_for_software_updates(flash_message=False):
    """Check for a new release of Invenio.

    :return: True if you have latest version, else False if you need to upgrade
             or None if server was not reachable.
    """
    from invenio_base.globals import cfg
    from invenio_base.i18n import _
    try:
        find = re.compile('Invenio v[0-9]+.[0-9]+.[0-9]+(\-rc[0-9])?'
                          ' is released')

        release_notes = 'https://raw.githubusercontent.com/' \
            'inveniosoftware/invenio/master/RELEASE-NOTES'

        webFile = urllib.request.urlopen(release_notes)

        temp = ""
        version = ""
        version1 = ""
        while True:
            temp = webFile.readline()
            match1 = find.match(temp)
            try:
                version = match1.group()
                break
            except Exception:
                pass
            if not temp:
                break

        webFile.close()
        submatch = re.compile('[0-9]+.[0-9]+.[0-9]+(\-rc[0-9])?')
        version1 = submatch.search(version)
        web_version = version1.group().split(".")

        local_version = cfg['CFG_VERSION'].split(".")

        if (web_version[0] > local_version[0] or
                web_version[0] == local_version[0] and
                web_version[1] > local_version[1] or
                web_version[0] == local_version[0] and
                web_version[1] == local_version[1] and
                web_version[2] > local_version[2]):
            if flash_message:
                flash(_('A newer version of Invenio is available for '
                        'download. You may want to visit '
                        '<a href="%(wiki)s">%()s</a>',
                        wiki='<a href=\"http://invenio-software.org/wiki/'
                             '/Installation/Download'), 'warning(html_safe)')

            return False
    except Exception as e:
        print(e)
        if flash_message:
            flash(_('Cannot download or parse release notes '
                    'from %(release_notes)s', release_notes=release_notes),
                  'error')
        return None
    return True


class Manager(FlaskExtManager):

    """Custom manager implementation with signaling support."""

    def add_command(self, name, command):
        """Wrap default ``add_command`` method."""
        sender = command.run if isinstance(command.run, FunctionType) \
            else command.__class__

        class SignalingCommand(command.__class__):

            def __call__(self, *args, **kwargs):
                app = self.app if not len(args) else args[0]
                with app.test_request_context():
                    pre_command.send(sender, args=args, **kwargs)
                res = super(SignalingCommand, self).__call__(*args, **kwargs)
                with app.test_request_context():
                    post_command.send(sender, args=args, **kwargs)
                return res

        command.__class__ = SignalingCommand
        return super(Manager, self).add_command(name, command)


def set_serve_static_files(sender, *args, **kwargs):
    """Enable serving of static files for `runserver` command.

    Normally Apache serves static files, but during development and if you are
    using the Werkzeug standalone development server, you can set this flag to
    `True`, to enable static file serving.
    """
    current_app.config.setdefault('CFG_FLASK_SERVE_STATIC_FILES', True)

pre_command.connect(set_serve_static_files, sender=Server)


def create_ssl_context(config):
    """Create :class:`ssl.SSLContext` from application config.

    :param config: Dict-like application configuration.
    :returns: A valid context or in case TLS is not enabled `None`.

    The following configuration variables are processed:

    ============================ ==============================================
    `SERVER_TLS_ENABLE`          If `True`, a SSL context will be created. In
                                 this case, the required configuration
                                 variables must be provided.
    `SERVER_TLS_KEY` (required)  Filepath (string) of private key provided as
                                 PEM file.
    `SERVER_TLS_CERT` (required) Filepath (string) of your certificate plus
                                 all intermediate certificate, concatenated in
                                 that order and stored as PEM file.
    `SERVER_TLS_KEYPASS`         If private key is encrypted, a password can be
                                 provided.
    `SERVER_TLS_PROTOCOL`        String that selects a protocol from
                                 `ssl.PROTOCOL_*`. Defaults to `SSLv23`. See
                                 :mod:`ssl` for details.
    `SERVER_TLS_CIPHERS`         String that selects possible ciphers according
                                 to the `OpenSSL cipher list format
                                 <https://www.openssl.org/docs/apps/
                                 ciphers.html>`_
    `SERVER_TLS_DHPARAMS`        Filepath (string) to parameters for
                                 Diffie-Helman key exchange. If not set the
                                 built-in parameters are used.
    `SERVER_TLS_ECDHCURVE`       Curve (string) that should be used for
                                 Elliptic Curve-based Diffie-Helman key
                                 exchange. If not set, the defaults provided by
                                 OpenSSL are used.
    ============================ ==============================================

    .. note:: In case `None` is returned because of a non-enabling
        configuration, TLS will be disabled. It is **not** possible to have a
        TLS and non-TLS configuration at the same time. So if TLS is activated,
        no non-TLS connection are accepted.

    from invenio_utils.. important:: Keep in mind to change `CFG_SITE_URL` and
        `CFG_SITE_SECURE_URL` according to your TLS configuration. This does
        not only include the protocol (`http` vs `https`) but also the hostname
        that has to match the common name in your certificate. If a wildcard
        certificate is provided, the hostname stated in
        `CFG_SITE[_SECURE]_URL` must match the wildcard pattern.

    """
    ssl_context = None

    if config.get('SERVER_TLS_ENABLE', False):
        if 'SERVER_TLS_KEY' not in config \
                or 'SERVER_TLS_CERT' not in config:
            raise AttributeError(
                '`SERVER_TLS_KEY` and `SERVER_TLS_CERT` required!'
            )

        # CLIENT_AUTH creates a server context, so do not get confused here
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)

        if 'SERVER_TLS_PROTOCOL' in config:
            ssl_context.protocol = getattr(
                ssl,
                'PROTOCOL_{}'.format(config.get('SERVER_TLS_PROTOCOL'))
            )

        ssl_context.load_cert_chain(
            certfile=config.get('SERVER_TLS_CERT'),
            keyfile=config.get('SERVER_TLS_KEY'),
            password=config.get('SERVER_TLS_KEYPASS', None)
        )
        if 'SERVER_TLS_CIPHERS' in config:
            ssl_context.set_ciphers(
                config.get('SERVER_TLS_CIPHERS')
            )
        if 'SERVER_TLS_DHPARAMS' in config:
            ssl_context.load_dh_params(
                config.get('SERVER_TLS_DHPARAMS')
            )
        if 'SERVER_TLS_ECDHCURVE' in config:
            ssl_context.set_ecdh_curve(
                config.get('SERVER_TLS_ECDHCURVE')
            )

        # that one seems to be required for werkzeug
        ssl_context.check_hostname = False
    return ssl_context


def register_manager(manager):
    """Register all manager plugins and default commands with the manager."""
    from six.moves.urllib.parse import urlparse
    managers = RegistryProxy('managers', ModuleAutoDiscoveryRegistry, 'manage')

    def extract_name(name):
        """Guess manager name."""
        parts = name.split('.')
        if len(parts) == 2:
            return parts[0].split('_')[-1]
        return parts[-2]

    with manager.app.app_context():
        for script in find_modules('invenio_base.scripts'):
            manager.add_command(script.split('.')[-1],
                                import_string(script + ':manager'))
        for script in managers:
            if script.__name__ == 'invenio_base.manage':
                continue
            manager.add_command(extract_name(script.__name__),
                                getattr(script, 'manager'))

    manager.add_command("clean", Clean())
    manager.add_command("show-urls", ShowUrls())
    manager.add_command("shell", Shell())

    parsed_url = urlparse(manager.app.config.get('CFG_SITE_URL'))
    host = manager.app.config.get(
        'SERVER_BIND_ADDRESS',
        parsed_url.hostname or '127.0.0.1'
    )
    port = manager.app.config.get(
        'SERVER_BIND_PORT',
        parsed_url.port or 80
    )

    ssl_context = create_ssl_context(manager.app.config)

    runserver = Server(host=host, port=port, ssl_context=ssl_context)
    manager.add_command("runserver", runserver)

    # FIXME separation of concerns is violated here.
    from invenio_ext.collect import collect
    collect.init_script(manager)

    from invenio_ext.assets import command, bower
    manager.add_command("assets", command)
    manager.add_command("bower", bower)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2013, 2014, 2015 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
    OneDrive file system
    --------------------

    Known issues:

    * Flush and close, both call write contents and because of that
      the file on cloud is overwrite twice...
"""

import os
import time
from UserDict import UserDict

import six
# python filesystem imports
from fs.base import FS
from fs.errors import CreateFailedError, NoPathURLError, \
    OperationFailedError, RemoteConnectionError, ResourceInvalidError, \
    ResourceNotFoundError, UnsupportedError
from fs.filelike import SpooledTemporaryFile
from fs.remote import RemoteFileBuffer
from onedrive import api_v5

# Items in cache are considered expired after 5 minutes.
CACHE_TTL = 300
# Max size for spooling to memory before using disk (5M).
MAX_BUFFER = 1024**2 * 5


class CacheItem(object):
    """Represents a path in the cache. There are two components to a path.
       It's individual metadata, and the children contained within it."""

    def __init__(self, metadata=None, children=None, timestamp=None,
                 parent=None):
        self.metadata = metadata
        self.children = children
        self.parent = parent
        if timestamp is None:
            timestamp = time.time()
        self.timestamp = timestamp

    def add_child(self, name, client=None):
        if self.children is None:
            if client is not None:
                # This is a fix. When you add a child to a folder that
                # was still not listed, that folder gets only one
                # child when you list it afterwards. So this fix
                # first tries to check are the files/folders inside
                # this directory on cloud.
                client.children(self.metadata['id'])
            else:
                self.children = [name]
        else:
            if name not in self.children:
                self.children.append(name)

    def del_child(self, name):
        if self.children is None:
            return
        try:
            i = self.children.index(name)
        except ValueError:
            return
        self.children.pop(i)

    def _get_expired(self):
        if self.timestamp <= time.time() - CACHE_TTL:
            return True
    expired = property(_get_expired)

    def renew(self):
        self.timestamp = time.time()


class OneDriveCache(UserDict):

    def __init__(self, client):
        self._client = client
        UserDict.__init__(self)

    def set(self, path, metadata, children=None, parent=None):
        self[path] = CacheItem(metadata, children=children, parent=parent)
        if parent is not None:
            if parent in self:
                self.get(parent).add_child(path, self._client)

    def pop(self, path, default=None):
        value = UserDict.pop(self, path, default)
        if value.parent is not None:
            if value.parent in self:
                self.get(value.parent).del_child(value.metadata['id'])
        return value


class OneDriveClient(api_v5.OneDriveAPI):

    def __init__(self, credentials):
        self.cache = OneDriveCache(self)
        self.auth_access_token = credentials.get("access_token", None)
        self.auth_refresh_token = credentials.get("refresh_token", None)
        self.auth_redirect_uri = credentials.get("redirect_uri", None)
        self.auth_scope = credentials.get("scope", None)
        self.client_id = credentials.get("client_id", None)
        self.client_secret = credentials.get("client_secret", None)

    def metadata(self, path):
        "Gets metadata for a given path."
        item = self.cache.get(path)
        if not item or item.metadata is None or item.expired:
            try:
                metadata = super(OneDriveClient, self).info(path)
            except api_v5.ProtocolError as e:
                if e.code == 404:
                    raise ResourceNotFoundError(path)
                raise OperationFailedError(opname='metadata', path=path,
                                           msg=str(e))
            except:
                raise RemoteConnectionError(
                    "Most probable reasons: access token has expired "
                    "or user credentials are invalid.")
            item = self.cache[path] = CacheItem(metadata)
        # Copy the info so the caller cannot affect our cache.
        return dict(item.metadata.items())

    def children(self, path):
        "Gets children of a given path."
        update = False
        item = self.cache.get(path)
        if item:
            if item.expired:
                update = True
            else:
                if item.metadata["type"] != "folder" and \
                        not ("folder" in path):
                    raise ResourceInvalidError(path)
            if not item.children:
                update = True
        else:
            update = True
        if update:
            try:
                metadata = super(OneDriveClient, self).info(path)
                if metadata["type"] != "folder" and not ("folder" in path):
                    raise ResourceInvalidError(path)
                children = []
                contents = super(OneDriveClient, self).listdir(path)
                for child in contents:
                    children.append(child['id'])
                    self.cache[child['id']] = CacheItem(child, parent=path)
                item = self.cache[path] = CacheItem(metadata, children)
            except api_v5.ProtocolError as e:
                if e.code == 404:
                    raise ResourceNotFoundError(path)
                if not item or e.resp.status != 304:
                    raise OperationFailedError(opname='metadata', path=path,
                                               msg=str(e))
                # We have an item from cache (perhaps expired), but it's
                # hash is still valid (as far as OneDrive is concerned),
                # so just renew it and keep using it.
                item.renew()
            except:
                raise RemoteConnectionError(
                    "Most probable reasons: access token has expired "
                    "or user credentials are invalid.")
        return item.children

    def file_create_folder(self, parent_id, title):
        "Add newly created directory to cache."
        try:
            metadata = super(OneDriveClient, self).mkdir(title, parent_id)
        except api_v5.ProtocolError as e:
            if e.code == 405:
                raise ResourceInvalidError(parent_id)
            if e.code == 404:
                raise ResourceNotFoundError(parent_id)
            raise OperationFailedError(opname='file_create_folder', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")

        self.cache.set(metadata["id"], metadata, parent=parent_id)
        return metadata['id']

    def file_copy(self, src, dst):
        try:
            metadata = super(OneDriveClient, self).copy(src, dst, False)
        except api_v5.ProtocolError as e:
            if e.code == 404:
                raise ResourceNotFoundError(
                    "Parent or source file don't exist")
            raise OperationFailedError(opname='file_copy', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")
        self.cache.set(metadata['id'], metadata, parent=dst)
        return metadata['id']

    def file_move(self, src, dst):
        try:
            metadata = super(OneDriveClient, self).copy(src, dst, True)
        except api_v5.ProtocolError as e:
            if e.code == 404:
                raise ResourceNotFoundError(
                    "Parent or source file don't exist")
            raise OperationFailedError(opname='file_move', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")

        self.cache.pop(src, None)
        self.cache.set(metadata['id'], metadata, metadata['parent_id'])
        return metadata['id']

    def file_delete(self, path):
        try:
            super(OneDriveClient, self).delete(path)
        except api_v5.ProtocolError as e:
            if e.code == 404:
                raise ResourceNotFoundError(path)
            raise OperationFailedError(opname='file_delete', msg=str(e))
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")
        self.cache.pop(path, None)

    def put_file(self, parent_id, title, content, overwrite=False):
        try:
            metadata = super(OneDriveClient, self).put(
                (title, content), parent_id, overwrite=overwrite)
        except api_v5.ProtocolError as e:
            if e.code == 404:
                raise ResourceNotFoundError(parent_id)
            raise OperationFailedError(opname='put_copy', msg=str(e))
        except TypeError as e:
            raise ResourceInvalidError("put_file")
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")
        return metadata['id']

    def get_file(self, path):
        metadata = None

        if not self.cache.get(path, None):
            try:
                metadata = super(OneDriveClient, self).info(path)
            except api_v5.ProtocolError as e:
                if e.code == 404:
                    raise ResourceNotFoundError("Source file doesn't exist")

                raise OperationFailedError(opname='get_file', msg=str(e))
            except:
                raise RemoteConnectionError(
                    "Most probable reasons: access token has expired "
                    "or user credentials are invalid.")
            self.cache.set(metadata['id'], metadata,
                           parent=metadata['parent_id'])
        else:
            item = self.cache[path]
            metadata = item.metadata

        return super(OneDriveClient, self).get(path)

    def update_file(self, file_id, new_file_info):
        try:
            metadata = super(OneDriveClient, self).info_update(
                file_id, new_file_info)
        except api_v5.ProtocolError as e:
            if e.resp.status == 404:
                raise ResourceNotFoundError(path=file_id)

            raise OperationFailedError(opname='update_file', msg=e.resp.reason)
        except:
            raise RemoteConnectionError(
                "Most probable reasons: access token has expired "
                "or user credentials are invalid.")
        self.cache.pop(file_id, None)
        self.cache.set(metadata['id'], metadata, parent=metadata['parent_id'])
        return metadata['id']


class OneDriveFS(FS):
    """
        Sky drive file system
    """
    __name__ = "OneDrive"

    _meta = {'thread_safe': True,
             'virtual': False,
             'read_only': False,
             'unicode_paths': True,
             'case_insensitive_paths': False,
             'network': True,
             'atomic.move': True,
             'atomic.copy': True,
             'atomic.makedir': True,
             'atomic.rename': False,
             'atomic.setconetns': True
             }

    def __init__(self, root=None, credentials=None, thread_synchronize=True):
        self._root = root
        self._credentials = credentials
        self.cached_files = {}

        if self._root is None or self._root == "/" or self._root == "":
            self._root = "me/skydrive"

        if self._credentials is None:
            if "ONEDRIVE_ACCESS_TOKEN" not in os.environ:
                raise CreateFailedError(
                    "ONEDRIVE_ACCESS_TOKEN is not set in os.environ")
            else:
                self._credentials['access_token'] = os.environ.get(
                    'ONEDRIVE_ACCESS_TOKEN')
                self._credentials['refresh_token'] = os.environ.get(
                    'ONEDRIVE_REFRESH_TOKEN', None)
                self._credentials['redirect_uri'] = os.environ.get(
                    'ONEDRIVE_REDIRECT_URI', None)
                self._credentials['client_id'] = os.environ.get(
                    'ONEDRIVE_CLIENT_ID', None)
                self._credentials['client_secret'] = os.environ.get(
                    'ONEDRIVE_CLIENT_SECRET', None)

        self.client = OneDriveClient(self._credentials)
        super(OneDriveFS, self).__init__(thread_synchronize=thread_synchronize)

    def __repr__(self):
        args = (self.__class__.__name__, self._root)
        return '<FileSystem: %s - Root Directory: %s>' % args

    __str__ = __repr__

    def __unicode__(self):
        args = (self.__class__.__name__, self._root)
        return u'<FileSystem: %s - Root Directory: %s>' % args

    def _update(self, path, data):
        """
        Updates contents of an existing file

        @param path: Id of the file for which to update content
        @param data: Contents to write to the file
        @return: Id of the updated file
        """
        path = self._normpath(path)

        if isinstance(data, six.string_types):
            string_data = data
        else:
            try:
                data.seek(0)
                string_data = data.read()
            except:
                raise ResourceInvalidError("Unsupported type")

        f = self.getinfo(path)
        return self.client.put(
            (f["title"], string_data), f["parent_id"], True)['id']

    def setcontents(self, path, data="", chunk_size=64 * 1024, **kwargs):
        """
        Sets new content to remote file

        Method works only with existing files and sets new content to them.
        @param path: Id of the file in which to write the new content
        @param contents: File contents as a string, or any object with read
            and seek methods
        @param kwargs: additional parameters like:
            encoding: the type of encoding to use if data is text
            errors: encoding errors
        @param chunk_size: Number of bytes to read in a chunk, if the
            implementation has to resort to a read / copy loop.
        @return: Id of the updated file
        """
        path = self._normpath(path)

        encoding = kwargs.get("encoding", None)
        errors = kwargs.get("errors", None)

        if isinstance(data, six.text_type):
            data = data.encode(encoding=encoding, errors=errors)

        return self._update(path, data)

    def createfile(self, path, wipe=True, **kwargs):
        """
        Creates always an empty file, even if another file with the same name
        exists.

        @param path: path to the new file. It has to be in one of following forms:
            - parent_id/file_title.ext
            - file_title.ext or /file_title.ext - In this cases root directory is the parent
        @param wipe: New file with empty content. If a file with the same name exists it will be
            overwritten.
        @raise ResourceNotFoundError: If parent doesn't exist.
        @attention: Root directory is the current root directory of this instance of
            filesystem and not the root of your Google Drive.
        @return: Id of the created file
        """
        parts = path.split("/")
        if parts[0] == "":
            parent_id = self._root
            title = parts[1]
        elif len(parts) == 2:
            parent_id = parts[0]
            title = parts[1]
            if not self.exists(parent_id):
                raise ResourceNotFoundError("parent doesn't exist")
        else:
            parent_id = self._root
            title = parts[0]

        return self.client.put_file(parent_id, title, "", wipe)

    def open(self, path, mode='r', buffering=-1, encoding=None,
             errors=None, newline=None, line_buffering=False, **kwargs):
        """
        Open the named file in the given mode.

        This method downloads the file contents into a local temporary file
        so that it can be worked on efficiently.  Any changes made to the
        file are only sent back to cloud storage when the file is flushed or closed.
        @param path: Id of the file to be opened
        @param mode: In which mode to open the file

        @raise ResourceNotFoundError: If given path doesn't exist and 'w' is not in mode
        @return: RemoteFileBuffer object
        """
        path = self._normpath(path)

        spooled_file = SpooledTemporaryFile(mode=mode, bufsize=MAX_BUFFER)

        #  Truncate the file if requested
        if "w" in mode:
            try:
                self._update(path, "")
            except:
                path = self.createfile(path, True)
        else:
            try:
                spooled_file.write(self.client.get_file(path))
                spooled_file.seek(0, 0)
            except Exception as e:
                if "w" not in mode and "a" not in mode:
                    raise ResourceNotFoundError("%r" % e)
                else:
                    path = self.createfile(path, True)

        return RemoteFileBuffer(self, path, mode, spooled_file)

    def is_root(self, path):
        path = self._normpath(path)
        if path == self._root:
            return True
        else:
            return False

    def copy(self, src, dst, overwrite=False, chunk_size=1024 * 64):
        """
        @param src: Id of the file to be copied
        @param dst: Id of the folder in which to copy the file
        @param overwrite: This is never true for OneDrive
        @return: Id of the copied file
        """
        return self.client.file_copy(src, dst)

    def copydir(self, src, dst, overwrite=False, ignore_errors=False,
                chunk_size=16384):
        """
        NOTE: OneDrive doesn't support copy of folders. And to implement it
              over copy method will be very inefficient
        """
        raise NotImplemented("If implemented method will be very inefficient")

    def rename(self, src, dst):
        """
        @param src: Id of the file to be renamed
        @param dst: New title of the file
        @raise UnsupportedError: If trying to rename the root directory
        @return: Id of the renamed file
        """
        if self.is_root(path=src):
            raise UnsupportedError("Can't rename the root directory")

        return self.client.update_file(src, {"name": dst})

    def remove(self, path):
        """
        @param path: id of the folder to be deleted
        @return: None if removal was successful
        """
        path = self._normpath(path)
        if self.is_root(path=path):
            raise UnsupportedError("Can't remove the root directory")
        if self.isdir(path=path):
            raise ResourceInvalidError("Specified path is a directory")

        return self.client.file_delete(path)

    def removedir(self, path):
        """
        @param path: id of the folder to be deleted
        @return: None if removal was successful
        """
        path = self._normpath(path)

        if not self.isdir(path):
            raise ResourceInvalidError("Specified path is a directory")
        if self.is_root(path=path):
            raise UnsupportedError("remove the root directory")

        return self.client.file_delete(path)

    def makedir(self, path, recursive=False, allow_recreate=False):
        """
        @param path: path to the folder you want to create.
            it has to be in one of the following forms:
                - parent_id/new_folder_name  (when recursive is False)
                - parent_id/new_folder1/new_folder2...  (when recursive is True)
                - /new_folder_name to create a new folder in root directory
                - /new_folder1/new_folder2... to recursively create a new folder in root
        @param recursive: allows recursive creation of directories
        @param allow_recreate: for OneDrive this param is always False, it will
            never recreate a directory
        """
        parts = path.split("/")

        if parts[0] == "":
            parent_id = self._root
        elif len(parts) >= 2:
            parent_id = parts[0]
            if not self.exists(parent_id):
                raise ResourceNotFoundError(
                    "parent with the id '%s' doesn't exist" % parent_id)

        if len(parts) > 2:
            if recursive:
                for i in range(len(parts) - 1):
                    title = parts[i + 1]
                    resp = self.client.file_create_folder(parent_id, title)
                    parent_id = resp
            else:
                raise UnsupportedError("recursively create a folder")
            return resp
        else:
            if len(parts) == 1:
                title = parts[0]
                parent_id = self._root
            else:
                title = parts[1]
            return self.client.file_create_folder(parent_id, title)

    def move(self, src, dst, overwrite=False, chunk_size=16384):
        """
        @param src: id of the file to be moved
        @param dst: id of the folder in which the file will be moved
        @param overwrite: for Sky drive it is always false
        @param chunk_size: if using chunk upload
        @return: Id of the moved file
        """
        if self.isdir(src):
            raise UnsupportedError("move a directory")
        return self.client.file_move(src, dst)

    def movedir(self, src, dst, overwrite=False, ignore_errors=False,
                chunk_size=16384):
        """
        @attention: onedrive API doesn't allow to move folders
        """
        raise UnsupportedError("move a directory")

    def isdir(self, path):
        """
        Checks if the given path is a folder

        @param path: id of the object to check
        @attention: this method doesn't check if the given path exists
            it will return true or false even if the file/folder doesn't exist
        """
        path = self._normpath(path)
        info = self.client.info(path)
        return "folder" in path or info['type'] == "folder"

    def isfile(self, path):
        """
        Checks if the given path is a file

        @param path: id of the object to check
        @attention: this method doesn't check if the given path exists
            it will return true or false even if the file/folder doesn't exist
        """
        path = self._normpath(path)
        info = self.client.info(path)
        return "file" in path or info['type'] == "file"

    def exists(self, path):
        """
        Checks if a the specified path exists

        @param path: Id of the file/folder to check
        """
        try:
            return self.client.metadata(path)
        except RemoteConnectionError as e:
            raise RemoteConnectionError(e)
        except:
            return False

    def listdir(self, path=None, wildcard=None, full=False, absolute=False,
                dirs_only=False, files_only=False, overrideCache=False):
        """
        Lists the the files and directories under a given path.

        The directory contents are returned as a list of unicode paths.

        @param path: id of the folder to list
        @type path: string
        @param wildcard: Only returns paths that match this wildcard
        @type wildcard: string containing a wildcard, or a callable that accepts a path and returns a boolean
        @param full: returns full paths (relative to the root)
        @type full: bool
        @param absolute: returns absolute paths (paths beginning with /)
        @type absolute: bool
        @param dirs_only: if True, only return directories
        @type dirs_only: bool
        @param files_only: if True, only return files
        @type files_only: bool
        @return: a list of unicode paths
        """
        path = self._normpath(path)
        flist = self.client.children(path)
        dirContent = self._listdir_helper('', flist, wildcard, full, absolute,
                                          dirs_only, files_only)

        return dirContent

    def listdirinfo(self, path=None, wildcard=None, full=False, absolute=False,
                    dirs_only=False, files_only=False):
        """
        Retrieves a list of paths and path info under a given path.

        This method behaves like listdir() but instead of just returning
        the name of each item in the directory, it returns a tuple of the
        name and the info dict as returned by getinfo.

        @param path: id of the folder
        @param wildcard: filter paths that match this wildcard
        @param dirs_only: only retrieve directories
        @type dirs_only: bool
        @param files_only: only retrieve files
        @type files_only: bool
        @return: tuple of the name and the info dict as returned by getinfo.
        """
        # Optimised listdir from pyfs
        return [(p, self.getinfo(p)) for p in self.listdir(
            path, wildcard=wildcard, full=full, absolute=absolute,
            dirs_only=dirs_only, files_only=files_only)]

    def getinfo(self, path):
        """
        Returned information is metadata from cloud service +
            a few more fields with standard names for some parts
            of the metadata.
        @param path: file id for which to return informations
        @return: dictionary with informations about the specific file
        """
        path = self._normpath(path)

        # Fix for onedrive, because if you don't write the path in the way they
        # want. It will raise an exception with the error code 400.
        # That error means nothing because it's raised in many situations.
        if not self.exists(path):
            raise ResourceNotFoundError(path=path)
        return self._metadata_to_info(self.client.metadata(path))

    def getpathurl(self, path, allow_none=False):
        """Returns a url that corresponds to the given path, if one exists.

        If the path does not have an equivalent URL form (and allow_none is False)
        then a :class:`~fs.errors.NoPathURLError` exception is thrown. Otherwise the URL will be
        returns as an unicode string.

        @param path: object id for which to return url path
        @param allow_none: if true, this method can return None if there is no
            URL form of the given path
        @type allow_none: bool
        @raises `fs.errors.NoPathURLError`: If no URL form exists, and allow_none is False (the default)
        @rtype: unicode

        """
        url = None
        try:
            url = self.getinfo(path)['source']
        except RemoteConnectionError as e:
            raise RemoteConnectionError(e)
        except:
            if not allow_none:
                raise NoPathURLError(path=path)

        return url

    def desc(self, path):
        """
        @return: The title for the given path.
        """
        info = self.getinfo(path)
        return info["title"]

    def about(self):
        """This methods returns information about the current user with whose
        credentials is the file system instantiated."""
        user = self.client.info("me")
        info = {}
        quota = self.client.get_quota()
        info['cloud_storage_url'] = "https://onedrive.live.com/"
        info['user_name'] = user['first_name'].capitalize() + \
            " " + user['last_name'].capitalize()
        info['quota'] = 100 * (float(quota[1] - quota[0]) / float(quota[1]))
        return info

    def _normpath(self, path):
        """
        Method normalizes the path for sky drive.
        @return: normalized path as a string
        """

        if path in ("/skydrive/camera_roll", "/skydrive/my_documents",
                    "/skydrive/my_photos", "/skydrive/public_documents",
                    "me/skydrive"):
            return path
        elif path == self._root:
            return path
        elif path is None or path == "":
            return self._root
        elif len(path.split("/")) > 2:
            return path.split("/")[-1]
        elif path[0] == "/" and len(path) == 1:
            return self._root
        elif path[0] == "/":
            return path[1:]
        elif len(path) == 0:
            return self._root

        return path

    def _metadata_to_info(self, metadata, localtime=False):
        """
        Method adds a few standard names to the metadata:
            size - the size of the file/folder
            isdir - is something a file or a directory
            created_time - the time of the creation
            path - path to the object which metadata are we showing
            revision - sky drive doesn't have a revision parameter
            modified - time of the last modification
        @return: The full metadata and a few more fields
            with standard names.
        """
        path = metadata.get('id', '0')
        isdir = "folder" in path or metadata.get('type') == "folder"
        info = {
            'isdir': isdir,
            'created_time': metadata.get('createdDate', 0),
            'title': metadata.get('name', 0),
            'path': path,
            'revision': None,
            'created_time': metadata.get('createdDate', 0),
            'modified': metadata.get("updated_time", 0)
        }
        info.update(metadata)

        return info
